import time
from loguru import logger as log
from cel.connectors.vapi.model.vapi_lead import VAPILead
from cel.gateway.model.base_connector import BaseConnector
from cel.connectors.telegram.model.telegram_attachment import TelegramAttachment
from cel.connectors.telegram.model.telegram_lead import TelegramLead
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.message import Message


class VAPIMessage(Message):
    
    def __init__(self, 
                 lead: ConversationLead, 
                 text: str = None, 
                 metadata: dict = None, 
                 date: int = None
                ):
        super().__init__(lead, text=text, date=date, metadata=metadata)
    
    
    def is_voice_message(self):
        #  check if the message has a voice attachment
        return False
    
    @classmethod
    async def load_from_message(cls, request, connector: BaseConnector = None):
        
        try: 
            messages = request.get("messages")
            if not messages:
                raise ValueError("No messages found in the message_dict")
            
            # take the last message
            user_message = messages[-1]
            
            if not user_message:
                raise ValueError("No message found in the message_dict")
            
            if user_message.get("role") != "user":
                raise ValueError("The message is not from a user")
            
            # get text from message or caption if it is a media message
            text = user_message.get("content")
            date = time.time()
            metadata = {'raw': user_message}
            lead = VAPILead.from_vapi_message(request, connector=connector)
            return VAPIMessage(lead=lead, text=text, date=date, metadata=metadata)
        except Exception as e:
            log.error(f"Error loading VAPI message from message_dict: {e}")
            return None        

    def __str__(self):
        return f"TelegramMessage: {self.text}"
        
    def __repr__(self):
        return f"TelegramMessage: {self.text}"
    
    


import time
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.conversation_peer import ConversationPeer


class VAPILead(ConversationLead):
    """ VAPILead class """

    def __init__(self, call_object: dict, **kwargs):
        """ VAPILead constructor 
        
        Args:
            call_object (dict): Call object
            Sample:
            {
                "id": "c7719e5c-ea98-40e1-b1dc-66131da31532",
                "orgId": "2ac97024-f9e9-425e-a846-ce5e2e3540f1",
                "createdAt": "2024-07-02T05:29:55.903Z",
                "updatedAt": "2024-07-02T05:29:55.903Z",
                "type": "webCall",  
                "status": "queued",
                "assistantId": "1d9d46ba-618e-4867-8797-5a8dc2f9f42x",
                "webCallUrl": "https://vapi.daily.co/E3pM5r6l7Q82gT4hElS7"
            }
        
        """
        super().__init__(**kwargs)
        self.call_object: dict = call_object


    def get_session_id(self):
        return f"{self.connector_name}:{self.call_object['id']}"  

    def to_dict(self):
        data = super().to_dict()
        data['call_object'] = self.call_object
        return data

    @classmethod
    def from_dict(cls, lead_dict):
        return VAPILead(
            call_object=lead_dict.get("call_object")
        )

    def __str__(self):
        return f"VAPILead: {self.call_object['id']}"
    
    
    @classmethod
    def from_vapi_message(cls, message: dict, **kwargs):
        call_object = message['call']
        metadata = {
            'date': time.time(),
            'raw': message
        }
        conversation_peer = ConversationPeer(
            name='Unknown',
            id=call_object['id'],
            phone=None,
            avatarUrl=None,
            email=None
        )
        return VAPILead(call_object=call_object, metadata=metadata, conversation_from=conversation_peer, **kwargs)


from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.outgoing.outgoing_message import OutgoingMessage, OutgoingMessageType
from cel.gateway.model.outgoing.outgoing_message_link import OutgoingLinkMessage
from cel.gateway.model.outgoing.outgoing_message_select import OutgoingSelectMessage
from cel.gateway.model.outgoing.outgoing_message_text import OutgoingTextMessage
from cel.gateway.model.outgoing.outgoing_message_buttons import OutgoingButtonsMessage

from loguru import logger as log

def outgoing_message_from_dict(data: dict) -> OutgoingMessage:
    
    try:
        """Creates an OutgoingMessage instance from a dictionary"""
        assert isinstance(data, dict),\
            "data must be a dictionary"
        assert "type" in data,\
            "data must have a 'type' key"
        
        if data["type"] == OutgoingMessageType.TEXT:
            return OutgoingTextMessage.from_dict(data)
        
        if data["type"] == OutgoingMessageType.SELECT:
            return OutgoingSelectMessage.from_dict(data)
        
        if data["type"] == OutgoingMessageType.LINK:
            return OutgoingLinkMessage.from_dict(data)
        
        if data["type"] == OutgoingMessageType.BUTTONS:
            return OutgoingButtonsMessage.from_dict(data)
        
        # TODO: add other message types here
        raise ValueError(f"Not implemented message type: {data['type']}")
        
    except AssertionError as e:
        log.error(f"Invalid outgoing message creating params: {e}")
        raise e
    except Exception as e:
        log.error(f"Error creating outgoing message: {e}")
        raise e
        
        

    


import asyncio
import hashlib
import json
from loguru import logger as log

def get_invariant_hash(obj):
    """
    Returns an MD5 hash of a JSON-serialized copy of the input object, with the 'metadata' field removed.
    This hash can be used to check if two objects are equivalent, regardless of their metadata.
    """
    
    # Copiar el objeto para evitar modificar el original
    obj_copy = obj.to_dict()

    # Eliminar el campo 'metadata' del objeto copiado
    if 'metadata' in obj_copy:
        del obj_copy['metadata']

    # Serializar el objeto en formato JSON
    json_str = json.dumps(obj_copy, sort_keys=True)

    # Generar un hash MD5 del JSON serializado
    md5_hash = hashlib.md5(json_str.encode('utf-8')).hexdigest()

    return md5_hash


def async_run(coro, then=None):
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:  # 'RuntimeError: There is no current event loop...'
            loop = None

        if loop and loop.is_running():
            log.debug('Async event loop already running. Adding coroutine to the event loop.')
            tsk = loop.create_task(coro)
            # ^-- https://docs.python.org/3/library/asyncio-task.html#task-object
            # Optionally, a callback function can be executed when the coroutine completes
            # tsk.add_done_callback(lambda t: print(f'Task done with result={t.result()}  << return val of main()'))
            tsk.add_done_callback(lambda t: then(t.result()))
            
        else:
            log.debug('Starting new event loop')
            result = asyncio.run(coro)
            
            



from abc import ABC
import time
from redis import asyncio as aioredis
import json
from dataclasses import asdict, dataclass
from cel.assistants.base_assistant import BaseAssistant
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.message import Message
from loguru import logger as log

Redis = aioredis.Redis
DEFUALT_MASTER_KEY = "123456"

@dataclass
class AuthEntry(ABC):
    id: str = None
    client_cmd_enabled: bool = False
    last_request: int = 0
    metadata: dict = None

class SessionMiddlewareEvents:
    new_conversation = "new_conversation"
    login = "login"
    logout = "logout"

class SessionMiddleware(ABC):
    """Middleware to secure client commands, session handling and user lyfe cycle related events
    Based on a Redis database.
    
    Triggered events:
        - new_conversation: Event called when a new conversation is started.
        - login: Event called when a user logs in.
        - logout: Event called when a user logs out.
        
    Message extended metadata:
        - time_since_last_request: Time since last request in seconds.
        
    """
    events: SessionMiddlewareEvents = SessionMiddlewareEvents()
    
    def __init__(self, redis: str | Redis = None, key_prefix: str = "authmw", master_key: str = None):
        self.client = redis if isinstance(redis, Redis) else aioredis.from_url(redis or 'redis://localhost:6379/0')
        self.key_prefix = key_prefix
        log.critical("No master key provided. Using default key") if not master_key else None
        self.master_key = master_key or DEFUALT_MASTER_KEY
        
    async def __call__(self, message: Message, connector: BaseConnector, assistant: BaseAssistant):
        assert isinstance(message, Message), "Message must be a Message object"
        assert isinstance(connector, BaseConnector), "Connector must be a BaseConnector object"
        
        entry = await self.get_entry(message.lead.get_session_id())
        
        # add last request time in message
        if entry:
            time_since_last_request = time.time() - (entry.last_request or 0)
            message.metadata = message.metadata or {}
            message.metadata['time_since_last_request'] = time_since_last_request
            await self.set_entry(message.lead.get_session_id(), client_cmd_enabled=entry.client_cmd_enabled, metadata=message.metadata)
        else:
            await self.set_entry(message.lead.get_session_id(), 
                                 client_cmd_enabled=False, 
                                 metadata={})
            await assistant.call_event(self.events.new_conversation, message.lead, message, connector)
            
        if not await self.__handle_login_command(message, connector, entry, assistant):
            return False
        
        if not await self.__secure_client_commands(message, connector, entry):
            return False
        
        return True
        
    
    async def __handle_login_command(self, 
                                     message: Message, 
                                     connector: BaseConnector, 
                                     entry: AuthEntry,
                                     assistant: BaseAssistant):
        text = message.text or ''
        if text.startswith("/logout"):
            await self.clear_auth(message.lead.get_session_id())
            await connector.send_text_message(message.lead, "You have been logged out!")
            await assistant.call_event(self.events.logout,  message.lead, message, connector)
            return False
        
        if text.startswith("/login"):
            parts = text.split(" ")
            if len(parts) == 2:
                if parts[1] == self.master_key:
                    await self.set_entry(message.lead.get_session_id(), client_cmd_enabled=True)
                    await assistant.call_event(self.events.login,  message.lead, message, connector)
                    await connector.send_text_message(message.lead, "You have been logged in.")
                    await connector.send_text_message(message.lead, "Welcome AI bender!")
                    return False
                else:
                    log.critical(f"Invalid master key provided for login")
                    return False
            else:
                log.critical(f"Invalid login command format")
                return False
        return True
        
    async def __secure_client_commands(self, message: Message, connector: BaseConnector, entry: AuthEntry):
        text = message.text or ''
        if text.startswith("/") and not text.startswith("/login") and not text.startswith("/logout"):
            if not entry or not entry.client_cmd_enabled:
                log.critical(f"Client command {text} from {message.lead.get_session_id()} is not allowed")
                return False

            
        if text.startswith("/reset all"):
            # clear entry for this session
            await connector.send_text_message(message.lead, "Resetting session data")
            await self.clear_auth(message.lead.get_session_id())
            
            
            
        return True
    
    async def set_entry(self, id: str, client_cmd_enabled: bool = False, metadata: dict = None):
        entry = AuthEntry(id=id, 
                          client_cmd_enabled=client_cmd_enabled, 
                          metadata=metadata,
                          last_request=time.time()
                        )
        await self.client.hset(self.key_prefix, id, json.dumps(asdict(entry)))
        return entry
        
    async def clear_auth(self, id: str):
        await self.client.hdel(self.key_prefix, id)
        
    async def get_entry(self, id: str):
        entry = await self.client.hget(self.key_prefix, id)
        if entry:
            entry = json.loads(entry)
            return AuthEntry(id=entry.get('id'), 
                            client_cmd_enabled=entry.get('client_cmd_enabled'), 
                            last_request=entry.get('last_request'),
                            metadata=entry.get('metadata'))

        return None



import asyncio
import inspect
import json
import re
from loguru import logger as log    
from typing import Callable, Dict, Optional
from cel.gateway.model.conversation_lead import ConversationLead



class PromptTemplate:
    """
    This class is used to compile a prompt template by replacing placeholders with values from the state.
    Placeholders are defined as {key} where key is a key in the state dictionary.
    The values in the state dictionary can be strings, integers, floats, lists, dictionaries, or functions that return
    any of the previous types.
    
    :param prompt: The prompt template with placeholders.
    :param initial_state: The initial state dictionary with values to replace placeholders. This state is static and
    will not be modified during the conversation. It will be used as a default value if the key is not found in the
    current stored state.
    """
    
    
    def __init__(self, prompt: str, initial_state: Optional[Dict] = {}):
        self.prompt = prompt
        self.initial_state = initial_state


    async def compile(self, state: Dict, lead: ConversationLead, message: str) -> str:
        """
        Compiles the prompt template by replacing placeholders with values from the state.
        
        :param state: The state dictionary with values to replace placeholders.
        :param lead: The ConversationLead instance.
        :param message: The message that triggered the prompt.
        :return: The compiled prompt as a string.
        """
        
        assert isinstance(state, dict), "PromptTemplate: state must be a dictionary"
        assert isinstance(lead, ConversationLead), "PromptTemplate: lead must be a ConversationLead instance"
        
        current_state = self.initial_state.copy()
        # merge initial state with state param
        current_state.update(state)
        
        async def compile_value(key):
            if key in current_state:
                value = current_state[key]
                try:
                    if callable(value):
                        return {key: await self.call_function(value, 
                                                              lead=lead, 
                                                              message=message,
                                                              current_state=state)}
                    return {key: str(value)}
                except Exception as e:
                    return {key: f"Error: {str(e)}"}
            return {key: ''}

        # Compile regex pattern outside the function if reused
        pattern = re.compile(r'\{(\w+)\}')
        tasks = [asyncio.create_task(compile_value(match)) for match in pattern.findall(self.prompt)]
        
        results = await asyncio.gather(*tasks)
        
        replacements = {key: value for result in results for key, value in result.items()}
        
        # Use a single pass to replace placeholders
        return pattern.sub(lambda m: replacements.get(m.group(1), ''), self.prompt)
            

    async def call_function(self, 
                            func: Callable, 
                            message: Optional[str] = None,
                            current_state: Optional[dict] = None,
                            lead: Optional[ConversationLead] = None) -> str:
        """
        Calls a provided function with dynamic arguments.

        :param func: The function to be called.
        :param message: Optional message to pass to the function.
        :param current_state: Optional current state to pass to the function.
        :param lead: Optional lead to pass to the function.
        :return: Function response as a JSON string if it's a dictionary or as a normal string.
        """
        
        if not lead:
            raise ValueError("PromptTemplate call_function: lead is required")
        
        if not isinstance(lead, ConversationLead):
            raise ValueError("PromptTemplate call_function: lead must be a ConversationLead instance")
        
        args_dict = {
            'lead': lead,
            'session_id': lead.get_session_id(),
            'message': message,
            'state': current_state
        }
        
        func_signature = inspect.signature(func)
        kwargs = {arg: args_dict[arg] for arg in func_signature.parameters if arg in args_dict}

        try:
            if inspect.iscoroutinefunction(func):
                response = await func(**kwargs)
            else:
                response = func(**kwargs)
        except Exception as e:
            # Log exception or handle it as needed
            response = f"Error occurred: {str(e)}"
            func_name = func.__name__ if hasattr(func, '__name__') else str(func)
            log.error(f"Error occurred while calling function {func_name}:  {str(e)}")
        
        return json.dumps(response) if isinstance(response, dict) else str(response)



import os
from geopy.geocoders import GoogleV3

GOOGLE_GEOCODING_API_KEY = os.getenv("GOOGLE_GEOCODING_API_KEY")


def googlev3_decode_location(location):
    if not GOOGLE_GEOCODING_API_KEY:
        raise ValueError("GOOGLE_GEOCODING_API_KEY env var is not set")
    
    geolocator = GoogleV3(api_key=GOOGLE_GEOCODING_API_KEY)
    location = geolocator.reverse(location)
    return location.address


if __name__ == "__main__":
    
    location = (40.748817, -73.985428)
    print(googlev3_decode_location(location))


    location = {
        "latitude": -34.391462,
        "longitude": -58.69314
    }
    
    location = (location["latitude"], location["longitude"])
    print(googlev3_decode_location(location))





from dataclasses import dataclass, field
from loguru import logger as log
import warnings
from cel.assistants.common import EventResponse
from cel.assistants.state_manager import AsyncStateManager
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.message import Message
from cel.assistants.base_assistant import BaseAssistant
from cel.gateway.model.outgoing.outgoing_message_text import OutgoingTextMessage
from cel.stores.history.base_history_provider import BaseHistoryProvider
from cel.stores.state.base_state_provider import BaseChatStateProvider


@dataclass
class Context:
    lead: ConversationLead
    connector: BaseConnector | None = None
    prompt: str | None = None
    message: Message | None = None
    assistant: BaseAssistant | None = None
    state: BaseChatStateProvider | None = None
    history: BaseHistoryProvider | None = None
    
    def state_manager(self, commit_on_error: bool = False):
        return AsyncStateManager(self.lead, self.state, commit_on_error)
    
    async def send_text_message(self, text: str, append_to_history: bool = True, is_private: bool = False):
        """ Send a direct message text to the user conversation. 
        This message will be sent without any AI processing.
        If append_to_history is True, the message will be added to the conversation history. 
        """
                
        # TODO:
        #  1- Create an OutgoingTextMessage object
        #  2- Send the message using MessageGateway in order to include middlewares in the loop
        #  3- Send the message using the connector
        #  4- Add the message to the history if append_to_history is True
        
        
        # 1- Create an OutgoingTextMessage object
        outgoing_message = OutgoingTextMessage(
            content=text,
            lead=self.lead,
            is_partial=False,
            is_private=is_private
        )
        
        
        try:
            from cel.gateway.message_gateway import MessageGateway
            from cel.gateway.message_gateway import StreamMode
            
            # 2- Send the message using MessageGateway in order to include middlewares in the loop
            mg = MessageGateway.instance()
            res = await mg.process_outgoing_msg_middlewares(
                    message=outgoing_message, 
                    is_partial=False, 
                    is_summary=False, 
                    mode=StreamMode.FULL)
            
            # 3- Send the message using the connector
            await self.lead.connector.send_message(outgoing_message)
            
            # 4- Add the message to the history if append_to_history is True
            if append_to_history:
                await self.assistant.append_message_to_history(self.lead, text)            

        except Exception as e:
            log.error(f"Error sending message: {e}")
            

            

            
            
    async def send_link_message(self, body: str, button_text: str, url: str, append_to_history: bool = False):
        # TODO: Add same logic as send_text_message!!!!!
        
        link = {"text": button_text, "url": url}
        await self.connector.send_link_message(self.lead, body, [link], is_partial=False)
        
        if append_to_history:
            await self.assistant.append_message_to_history(self.lead, f"{body} {url}")
    
    async def send_voice_message(self, text: str, append_to_history: bool = True):
        # TODO: Add same logic as send_text_message!!!!!
        """ Send a direct voice message to the user conversation. 
        This message will be sent without any AI processing. 
        And won't be stored in the conversation history. 
        Use this method only for messages that are not part of the conversation flow. """
        if self.connector:
            await self.connector.send_voice_message(self.lead, text, options={})
            
        if append_to_history:
            await self.assistant.append_message_to_history(self.lead, text)
            
    async def blend_message(self, text: str, history_length: int = None):
        # TODO: Add same logic as send_text_message!!!!!
        """ Blend a message into the conversation context. 
        and return the blended text. """    
        return await self.assistant.blend(self.lead, text, history_length)
        
    
    async def send_typing_action(self):
        """ Send a typing action to the user conversation. 
        Not all connectors support this feature. 
        It's fail safe, so if the connector does not support it, it will just ignore the action. """
        if self.connector:
            await self.connector.send_typing_action(self.lead)
    
    async def get_state(self):
        """ Get the current state of the conversation. """
        if self.state:
            return await self.state.get_store(self.lead.get_session_id())
        return None
    
    async def set_state(self, state: dict):
        """ Set the current state of the conversation. """
        if self.state:
            await self.state.set_store(self.lead.get_session_id(), state)
            
    async def get_history(self):
        """ Get the conversation history. """
        if self.history:
            return await self.history.get_history(self.lead.get_session_id())
        return None
    
    async def clear_history(self, message: Message):
        """ Clear the conversation history. """
        if self.history:
            await self.history.clear_history(self.lead.get_session_id())
            

    """ This method are meant to be used by used as shortcuts to create EventResponse objects. 
    EventResponse es a mechanism to delegate the response processing to the AI or to the Gateway.
    If the response require certain processing againts history (append meessage to history) or
    keep it hide from the user (is_private) and being added to the history.
    Also if the response should be blended into conversation context (blend) or not.   
    """
    async def cancel_ai_response(self):
        """ If you want to cancel further processing of the message by the AI """
        return EventResponse(disable_ai_response=True)
    
    @staticmethod
    def cancel_ai_response():
        """ If you want to cancel further processing of the message by the AI """
        return EventResponse(disable_ai_response=True)
    
    @staticmethod
    def response_text(text: str, disable_ai_response: bool = False, is_private: bool = False, blend: bool = False):
        warnings.warn(
            "The 'response_text' method is deprecated and will be removed in a future release.",
            DeprecationWarning,
            stacklevel=2
        )        
        return EventResponse(text=text, disable_ai_response=disable_ai_response, is_private=is_private, blend=blend)
    


import os
from elevenlabs import Voice, VoiceSettings, play
from elevenlabs.client import ElevenLabs
from cel.voice.base_voice_provider import BaseVoiceProvider


# https://help.elevenlabs.io/hc/en-us/articles/17883183930129-What-models-do-you-offer-and-what-is-the-difference-between-them

class ElevenLabsAdapter(BaseVoiceProvider):
    
    def __init__(self, api_key: str = None, 
                 default_voice: str = "N2lVS1w4EtoT3dr4eOWO", 
                 model: str = "eleven_turbo_v2_5",
                 settings: VoiceSettings = None,
                 preprocessing_callback: callable = None
                ):
        
        self.api_key = api_key
        self.default_voice = default_voice
        self.default_model = model
        self.default_voice_settings = settings or VoiceSettings(
                                                        stability=0.50, 
                                                        similarity_boost=0.75, 
                                                        style=0.0, 
                                                        use_speaker_boost=True
                                                    )
        
        key = api_key or os.getenv("ELEVENLABS_API_KEY")
        if not key:
            raise Exception("ELEVENLABS_API_KEY not found in the environment")
        
        self.client = ElevenLabs(api_key=key)
        self.preprocessing_callback = preprocessing_callback
        
        
    def TTS(self, text: str, voice: str = None, settings: VoiceSettings = None) -> Voice:
        
        if self.preprocessing_callback:
            text = self.preprocessing_callback(text)

        audio = self.client.generate(
            text=text,
            model=self.default_model,
            voice=Voice(
                voice_id=voice or self.default_voice,
                settings=settings or self.default_voice_settings,
            )
        )
        return audio
    
    
    def STT(self, audio: bytes | str) -> str:
        raise NotImplementedError
    
    
    
if __name__ == "__main__":
    
    import os
    import dotenv
    dotenv.load_dotenv()
    ELEVENLABS_API_KEY= os.getenv("ELEVENLABS_API_KEY")
    tts = ElevenLabsAdapter(
        api_key=ELEVENLABS_API_KEY,
        model=os.getenv("ELEVENLABS_MODEL", "eleven_turbo_v2_5"),
        preprocessing_callback=lambda x: x + " - from ElevenLabs"
    )
    audio = tts.TTS("This is a test")
    play(audio)



from abc import ABC, abstractmethod
import inspect
import json

from loguru import logger as log
from cel.assistants.function_response import FunctionResponse
from cel.gateway.model.base_connector import BaseConnector
from cel.assistants.common import EventResponse, FunctionDefinition
from cel.gateway.model.message import Message, ConversationLead
from cel.prompt.prompt_template import PromptTemplate
from cel.rag.providers.rag_retriever import RAGRetriever
from cel.stores.history.base_history_provider import BaseHistoryProvider
from cel.stores.state.base_state_provider import BaseChatStateProvider
from cel.stores.history.history_inmemory_provider import InMemoryHistoryProvider
from cel.stores.state.state_inmemory_provider import InMemoryStateProvider


class Events:
    START: str = "start"
    MESSAGE: str = "message"
    IMAGE: str = "image"
    AUDIO: str = "audio"
    END: str = "end"

class BaseAssistant(ABC):
    
    def __init__(self, 
                 name: str = None,
                 description: str = None,
                 prompt: PromptTemplate = None,
                 history_store: BaseHistoryProvider = None,
                 state_store: BaseChatStateProvider = None,
                ):
        self.name = name
        self.description = description
        self.function_handlers = {}
        self.event_handlers = {}
        self.client_commands_handlers = {}
        self.timeout_handlers = {}
        self._state_store = state_store or InMemoryStateProvider()
        self._history_store = history_store or InMemoryHistoryProvider()
        
        self.rag_retriever: RAGRetriever = None
        if prompt:
            assert isinstance(prompt, PromptTemplate), "prompt must be an instance of PromptTemplate"
        self.prompt = prompt or PromptTemplate('')

    def set_history_store(self, history_store: BaseHistoryProvider):
        # check type
        assert isinstance(history_store, BaseHistoryProvider), "history_store must be an instance of BaseHistoryProvider"
        self._history_store = history_store
        
    def set_state_store(self, state_store: BaseChatStateProvider):
        # check type
        assert isinstance(state_store, BaseChatStateProvider), "state_store must be an instance of BaseChatStateProvider"
        self._state_store = state_store
            
    def set_rag_retrieval(self, rag_retrieval: RAGRetriever):
        # check type
        assert isinstance(rag_retrieval, RAGRetriever), "rag_retrieval must be an instance of RAGRetriever"
        self.rag_retriever = rag_retrieval
        

    def function(self, name, desc, params):
        def decorator(func):
            self.function_handlers[name] = {
                # 'desc': desc,
                # 'params': params,
                'definition': FunctionDefinition(name=name, description=desc, parameters=params),
                'func': func
            }
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)          
                
            return wrapper
        return decorator    
        
        
    def event(self, name: Events | str):
        def decorator(func):
            self.event_handlers[name] = {
                'func': func
            }
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)          
                
            return wrapper
        return decorator
    
    
    def client_command(self, name):
        def decorator(func):
            self.client_commands_handlers[name] = {
                'func': func
            }
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)          
                
            return wrapper
        return decorator
    
    def timeout(self, label):
        def decorator(func):
            self.timeout_handlers[label] = {
                'func': func
            }
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)          
                
            return wrapper
        return decorator
    
    async def eval_client_command(self, command: str, args: list[str], message: Message):
        """Evaluate the client command and call the respective handler"""
        lead = message.lead
        connector = lead.connector
        
        if command in self.client_commands_handlers:
            func = self.client_commands_handlers[command]['func']
            # Build args_dict
            from cel.assistants.request_context import RequestContext
            ctx = RequestContext(lead=lead, 
                                 message=message,
                                 assistant=self,
                                 state=self._state_store,
                                 history=self._history_store,
                                 connector=connector)
            args_dict = {
                'session': lead.get_session_id(),
                'ctx': ctx,
                'lead': lead,
                'message': None,
                'connector': connector,
                'command': command,
                'args': args
            }
            
            # Build kwargs using args and args_dict
            args = inspect.getfullargspec(func).args
            kwargs = {}
            for arg in args:
                if arg in args_dict:
                    kwargs[arg] = args_dict[arg]

            # Check if the function is a coroutine (async function)
            if inspect.iscoroutinefunction(func):
                await func(**kwargs)
            else:
                func(**kwargs)      
            return False
        else:
            log.warning(f"Client command {command} not found in assistant implementation")
            return True
    
    async def call_event(self, event_name: str, lead: ConversationLead, message: Message = None, connector: BaseConnector = None, data: dict= None) -> EventResponse:
        """Call the respective event handler, if exists"""
        if event_name in self.event_handlers:
            func = self.event_handlers[event_name]['func']
            connector = connector or lead.connector
            from cel.assistants.request_context import RequestContext
            ctx = RequestContext(lead=lead, 
                                 message=message,
                                 assistant=self,
                                 state=self._state_store,
                                 history=self._history_store,
                                 connector=connector)
            
            # Build args_dict
            args_dict = {
                'session': lead.get_session_id(),
                'ctx': ctx,
                'lead': lead,
                'message': message,
                'connector': connector,
                'data': data,
                'state': self._state_store,
                'history': self._history_store
            }
            
            # Build kwargs using args and args_dict
            args = inspect.getfullargspec(func).args
            kwargs = {}
            for arg in args:
                if arg in args_dict:
                    kwargs[arg] = args_dict[arg]

            # Check if the function is a coroutine (async function)
            response = None
            if inspect.iscoroutinefunction(func):
                response = await func(**kwargs)
            else:
                response = func(**kwargs)    
                
            assert isinstance(response, EventResponse) or response is None, "Event handler must return an instance of EventResponse or None"
            return response
        else:
            # raise ValueError(f"Event {event_name} not found")
            log.error(f"Event {event_name} not found")
        
    
    async def call_function(self, func_name: str, params: dict, lead: ConversationLead) -> FunctionResponse:
        """Call the respective function handler, if exists. Map function arguments to the function handler signature"""
        from cel.assistants.function_context import FunctionContext
        connector = lead.connector
        if func_name in self.function_handlers:
            func = self.function_handlers[func_name]['func']
            definition = self.function_handlers[func_name]['definition']
            args = inspect.getfullargspec(func).args

            ctx = FunctionContext(lead=lead, 
                                  assistant=self,
                                  connector=connector, 
                                  state=self._state_store, 
                                  history=self._history_store,
                                  function_definition=definition,
                                  params=params)
            
            args_dict = {
                'session': lead.get_session_id(),
                'ctx': ctx,
                'params': params,
                'lead': lead,
                'message': None,
                'connector': connector,
                'state': self._state_store,
                'history': self._history_store                
            }

            # Build kwargs using args and args_dict
            kwargs = {}
            for arg in args:
                if arg in args_dict:
                    kwargs[arg] = args_dict[arg]

            # Check if the function is a coroutine (async function)
            response = None
            if inspect.iscoroutinefunction(func):
                response = await func(**kwargs)
            else:
                response = func(**kwargs)

            if not response:
                log.warning(f"Function {func_name} returned None")
                return FunctionResponse(text="Function executed. No response.")
                
            # assert isinstance(response, FunctionResponse) or response is None, "Function handler must return an instance of CommandResponse or None"
            if isinstance(response, FunctionResponse):
                return response

            if isinstance(response, EventResponse):
                return response
            
            # response is Any object?
            # Try to convert repsonse into a json string
            try:
                log.warning(f"Function {func_name} returned a non-FunctionResponse object. Converting to JSON.")                
                response_json = json.dumps(response)
                return FunctionResponse(text=response_json)
            except Exception as e:
                log.error(f"Function handler must return an instance of FunctionResponse or a JSON serializable object. Error: {e}")
                return FunctionResponse(text="Function response is not a valid JSON object")
        
        else:
            raise ValueError(f"Function {func_name} not found")
            
    @abstractmethod
    async def new_message(self, message: Message, local_state: dict = {}):
        raise NotImplementedError
    
    
    @abstractmethod
    async def blend(self, lead: ConversationLead, text: str, history_length: int = 3):
        raise NotImplementedError
    
    @abstractmethod
    async def process_client_command(self, lead: ConversationLead, command: str, args: list[str]):
        yield "Command not found"
        
    @abstractmethod
    async def append_message_to_history(self, lead: ConversationLead, message: str, role: str = "assistant"):
        raise NotImplementedError
    
    def get_functions(self) -> list[FunctionDefinition]:
       return [f['definition'] for f in self.function_handlers.values()] 

    def get_events(self):
        return self.event_handlers
    
    # str method
    def __str__(self):
        return f"{self.name}: {self.description}"
    
    # json dumps method
    def to_json(self):
        return {
            'name': self.name,
            'description': self.description
        }



import json
from loguru import logger as log

from langchain.load.load import load
from langchain.load.dump import dumps
from langchain_core.prompts import PromptTemplate as LangchainPromptTemplate
from langchain_openai import ChatOpenAI
from langsmith import traceable

from cel.assistants.base_assistant import BaseAssistant
from cel.assistants.router.utils import build_router_query
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.message import Message
from cel.stores.history.base_history_provider import BaseHistoryProvider
from cel.stores.state.base_state_provider import BaseChatStateProvider
from cel.stores.history.history_inmemory_provider import InMemoryHistoryProvider
from cel.stores.state.state_inmemory_provider import InMemoryStateProvider
from cel.assistants.macaw.macaw_history_adapter import MacawHistoryAdapter


class AgenticRouter(BaseAssistant):
    """
    The AgenticRouter class provides multi-assistant support for Cel.ai by routing incoming 
    messages to the most suitable assistant based on user intent. 
    It leverages a GPT-4o-based agentic router to determine the best assistant for handling 
    the user's request. To accurately detect user intent, the router considers a window 
    of messages from the history to provide context.

    If the router cannot correctly determine the user's intent, such as in the case 
    of the first message, it defaults to routing the message to the default assistant.

    It is important to note that this multi-assistant architecture uses shared history 
    and state storages to avoid inconsistencies and potential issues.

    Attributes:
        assistants (list[BaseAssistant]): A list of assistant instances to route messages to.
        history_store (BaseHistoryProvider): The provider for storing message history. Defaults to InMemoryHistoryProvider if not provided.
        state_store (BaseChatStateProvider): The provider for storing chat state. Defaults to InMemoryStateProvider if not provided.
        history_length (int): The number of historical messages to consider for context. Defaults to 5.
        llm: The language model used for intent detection. Defaults to None.
        default_assistant (int): The index of the default assistant to use when intent cannot be determined. Defaults to 0.
        
    Example:
    ```python
        # Create a list of assistants
        assistants = [build_balance_agent(), build_transfer_agent()]
        # Instantiate the Agentic Router
        ast = AgenticRouter(assistants=assistants) 
    ```

    Methods:
        infer_best_assistant(input_text: str) -> BaseAssistant:
            Determines the best assistant to handle the user's request based on the input text.

        build_dialog(lead: ConversationLead, text: str) -> list[dict]:
            Builds a dialog to be used by the agent to get the most suitable assistant to process the message.

        format_dialog_to_plain_text(dialog: list[dict]) -> str:
            Formats the dialog into plain text.

        get_assistant(lead: ConversationLead, text: str) -> BaseAssistant:
            Gets the assistant that best fits the input text.

        new_message(lead: ConversationLead, message: str, local_state: dict = {}) -> AsyncGenerator:
            Handles a new message by routing it to the appropriate assistant and yielding the response.
    """
        
    def __init__(self, 
                 assistants: list[BaseAssistant], 
                 history_store: BaseHistoryProvider = None,
                 state_store: BaseChatStateProvider = None,
                 history_length: int = 5,
                 llm=None,
                 default_assistant: int = 0):
        """
        Initializes the AgenticRouter instance.

        Args:
            assistants (list[BaseAssistant]): A list of assistant instances to route messages to.
            history_store (BaseHistoryProvider, optional): The provider for storing message history. Defaults to InMemoryHistoryProvider if not provided.
            state_store (BaseChatStateProvider, optional): The provider for storing chat state. Defaults to InMemoryStateProvider if not provided.
            history_length (int, optional): The number of historical messages to consider for context. Defaults to 5.
            llm (optional): The LLM model used for intent detection. Defaults to None.
            default_assistant (int, optional): The index of the default assistant to use when intent cannot be determined. Defaults to 0.

        Raises:
            AssertionError: If no assistants are provided or if the default assistant index is out of range.
        """
        
        
        super().__init__(name="Router Assistant", description="This assistant routes messages to other assistants")
        log.debug(f"Router Assistant created with {len(assistants)} assistants")
        assert len(assistants) > 0, "At least one assistant is required"
        assert default_assistant < len(assistants), "Default assistant index out of range"
        
        
        # List of assistants
        self._assistants = assistants
        self._current_assistant = self._assistants[default_assistant]
        self._hisory_length = history_length
        self._llm = llm
        
        # Init state and history store
        self._state_store = state_store or InMemoryStateProvider()
        self._history_store = history_store or InMemoryHistoryProvider()      
        
        # Make sure than all assistants share the same state and history store
        for ast in self._assistants:
            log.debug(f"Assistant Name: {ast.name} -> {ast.description}")
            
            if ast._state_store != self._state_store:
                # TODO: evaluate if this is the correct behavior
                # If the assistant has a different state store, overwrite it???
                log.critical(f"Assistant {ast.name} has different state store")
                ast.set_state_store(self._state_store)
                
            if ast._history_store != self._history_store:
                # If the assistant has a different history store, overwrite it
                # is important to keep the history store consistent
                # History consistency is important for the router to work properly
                log.warning(f"Assistant {ast.name} has different history store")
                ast.set_history_store(self._history_store)
                log.warning(f"Assistant {ast.name} history store overwritten!!")
            
            
        log.debug(f"Default Assistant: {self._current_assistant.name}")
    

    @traceable
    async def infer_best_assistant(self, input_text: str) -> BaseAssistant:
        """ Get the agent that best fits the input text """
        
        llm = self._llm or ChatOpenAI(model="gpt-4o", temperature=0, max_tokens=100)
        prompt_str = """From the following dialog, detect the user's intention, then return the most suitable assistant to handle the user's request: 
{input_text}

Available assistants:
Default Agent
{asts}

If you are not sure, please select the default agent.
Returns only the name of the assistant."""

        asts = "\n".join([f"{agent.name}" for i, agent in enumerate(self._assistants)])
        
        prompt = LangchainPromptTemplate.from_template(prompt_str)
        
        # invoke
        res = llm.invoke(prompt.format(input_text=input_text, asts=asts))
        ast_name = res.content
        
        if ast_name == "Default Agent":
            log.warning(f"AgenticRouter could not determine the assistant, using default assistant: {self._current_assistant.name}")
            return self._current_assistant
        
        # find the agent
        for i, agent in enumerate(self._assistants):
            if agent.name == ast_name:
                return agent
        
        raise ValueError(f"Agent {ast_name} not found")

    
    
    async def build_dialog(self, lead: ConversationLead, text: str):
        """ Build a dialog to be used by the agent to get the most suitable assistant to process the message """
        query = await build_router_query(self._history_store, lead, text, length=self._hisory_length)
        query.append({
            "role": "user",
            "text": text
        })
        return query
    
    async def format_dialog_to_plain_text(self, dialog: list[dict]):
        """ Format the dialog to plain text """
        return "\n".join([f"{d['role']}: {d['text']}" for d in dialog])
        
        
    async def get_assistant(self, lead: ConversationLead, text: str):
        """ Get the assistant that best fits the input text """
        dialog = await self.build_dialog(lead, text)
        plain_dialog = await self.format_dialog_to_plain_text(dialog)
        return await self.infer_best_assistant(plain_dialog)

    async def new_message(self, message: Message, local_state: dict = {}):
        log.debug(f"Router Assistant: new message: {message}")
        lead = message.lead
        ast = await self.get_assistant(lead, message.text)
        
        # Call 'message' event in the selected assistant
        res = await ast.call_event("message", lead, message)
        # assert res is None or isinstance(res, EventResponse)
        
        if res is None:
            log.debug(f"Event: 'message' response is None for message: {message.text}")
            
        # if res is not noe and res has attr disable_ai_response, and  res.disable_ai_response is True
        if res and hasattr(res, "disable_ai_response") and res.disable_ai_response:
            log.warning(f"Event 'message' response has disabled AI response for message: {message.text}")
            # Return empty response
            return
        
        assert isinstance(ast, BaseAssistant), "Agent must be a BaseAssistant instance"
        log.debug(f"Router Assistant selected: {ast.name}")
        # return 
    
        async for chunk in ast.new_message(message, local_state):
            yield chunk
        
       

    async def blend(self, lead: ConversationLead, text: str, history_length: int = None):
        log.debug(f"Router Assistant: blend: {text}")
        ast = await self.get_assistant(lead, text)
        
        return await ast.blend(lead, text, history_length)

        
    
    async def do_insights(self, lead: ConversationLead, targets: dict = {}, history_length: int = 10):
        log.warning("Router Assistant: do insights not implemented")
        return {}


    async def append_message_to_history(self, lead: ConversationLead, message: str, role: str = "assistant"):
        assert role in ["assistant", "user", "system"], "role must be one of: assistant, user, system"

        history = MacawHistoryAdapter(self._history_store)
        
        from langchain_core.messages import HumanMessage, SystemMessage, AIMessage
        entry = None
        if role == "assistant":
            entry = AIMessage(message)
        elif role == "user":
            entry = HumanMessage(message)
        elif role == "system":
            entry = SystemMessage(message)
            
        await history.append_to_history(lead, entry)
        
    
    async def process_client_command(self, lead: ConversationLead, command: str, args: list[str]):
    
        if command == "reset":
            
            if args and args[0] == "all":
                await self._history_store.clear_history(lead.get_session_id())
                yield "History cleared"
                await self._state_store.clear_store(lead.get_session_id())
                yield "State cleared"
                return

            await self._history_store.clear_history(lead.get_session_id())
            yield "History cleared"
            return
        
        if command == "state":
            state = await self._state_store.get_store(lead.get_session_id())
            if state is None:
                yield "No state found"
                return
            for k, v in state.items():
                yield f"{k}: {v}"
            return
        
        if command == "history":
            history = await self._history_store.get_history(lead.get_session_id()) or []    

            if history is None or len(history) == 0:
                yield "History is empty"
                return

            for h in history:
                aux = load(h)
                log.debug(f"History: {aux}")
                yield dumps(aux)
            
            return
        
        if command == "set":
            if len(args) < 2:
                yield "Not enough arguments"
                return
            key = args[0]
            value = args[1]
            state = self._state_store.get_store(lead.get_session_id())
            state[key] = value
            self._state_store.set_store(lead.get_session_id(), state)
            yield f"State updated: {key}: {value}"
            return
        
        if command == "prompt":
            yield "Macaw Assistant v0.1"

            if args and args[0] == "all":            
                prompt = self.prompt
                # split into 250 chars aprox chunks
                for i in range(0, len(prompt), 250):
                    yield prompt[i:i+250]
                return
            
            # first 250 chars
            yield self.prompt[:250]
            return

            
            


from abc import ABC
from .component import Component
import json
from .utils import del_none

class ListItem(Component):
    def __init__(self, title: str, id: str = None, description: str = None) -> None:
        self.title = title
        self.id = id or title
        self.description = description
        
    def __str__(self):
        item = {
            "id": self.id, 
            "title": self.title, 
            "description": self.description
        }
        del_none(item)
        return json.dumps(item)
        
    
    def __repr__(self):
        return f"ListItem(id={self.id}, title={self.title})"
        
    
    def __eq__(self, other):
        return self.id == other.id and self.title == other.title


if __name__ == "__main__":
    item = ListItem("id", "title")
    print(item)



from abc import ABC, abstractmethod


class KeyValueStore(ABC):
    
    @abstractmethod
    def get(self, key, callback=None):
        raise NotImplementedError()

    @abstractmethod
    def set(self, key, value, ttl=None):
        raise NotImplementedError()

    @abstractmethod
    def delete(self, key):
        raise NotImplementedError()
    
    @abstractmethod
    def clear(self):
        raise NotImplementedError()
    
    @abstractmethod
    def get_all(self):
        raise NotImplementedError()



class ListStore(ABC):
    
    @abstractmethod
    def list_append(self, key, value, ttl=None):
        raise NotImplementedError()
    
    @abstractmethod
    def list_clear(self, key):
        raise NotImplementedError()
    
    @abstractmethod
    def list_get(self, key):
        raise NotImplementedError()
    
    @abstractmethod
    def list_get_last(self, key, count: int):
        raise NotImplementedError()
    



from typing import List
from langchain_core.messages import BaseMessage, trim_messages
from cel.assistants.common import FunctionDefinition, Param
from langchain_core.messages import HumanMessage

def get_last_n_elements(msgs: List[BaseMessage], n: int) -> List[BaseMessage]:
    if msgs is None:
        return []
    
    if len(msgs) == 1 and msgs[0].type == "system":
        return msgs
     
    # if last message is not human. Add a mock human message
    # it seems that trim_messages param end_on=("human", "tool", "ai") is not working
    # as expected
    mock_message = HumanMessage("mock human message")
    if msgs[-1].type != "human":
        msgs.append(mock_message)
     
    res = trim_messages(
        msgs,
        # Keep the last <= n_count tokens of the messages.
        strategy="last",
        token_counter=len,
        # When token_counter=len, each message
        # will be counted as a single token.
        # Remember to adjust for your use case
        max_tokens=n,
        # Most chat models expect that chat history starts with either:
        # (1) a HumanMessage or
        # (2) a SystemMessage followed by a HumanMessage
        start_on=("human", "ai"),
        # Most chat models expect that chat history ends with either:
        # (1) a HumanMessage or
        # (2) a ToolMessage
        end_on=("human", "tool"),
        # Usually, we want to keep the SystemMessage
        # if it's present in the original history.
        # The SystemMessage has special instructions for the model.
        include_system=True,    
    )
    
    # Remove the mock message
    if mock_message in res:
        res.remove(mock_message)
        
    return res
            


def map_function_parameters_to_tool_message(parameters: list[Param]):
    assert isinstance(parameters, list),\
        "parameters must be a list"

    properties = {}
    for p in parameters:
        properties[p.name] = {
            "type": p.type,
            "description": p.description
        }
        if p.enum:
            properties[p.name]["enum"] = p.enum
            
    return properties


def map_function_to_tool_message(function: FunctionDefinition):
    assert isinstance(function, FunctionDefinition),\
        "function must be an instance of FunctionDefinition"
        
    return {
        "type": "function",
        "function": {
            "name": function.name,
            "description": function.description,
            "parameters": {
                "type": "object",
                "properties": map_function_parameters_to_tool_message(function.parameters),
                "required": [p.name for p in function.parameters if p.required]
            }
        }
    }

def map_functions_to_tool_messages(functions: list[FunctionDefinition]):
    assert isinstance(functions, list),\
        "functions must be a list"
        
    return [map_function_to_tool_message(f) for f in functions]




from abc import ABC
from dataclasses import dataclass
from loguru import logger as log
import chromadb
from cel.rag.stores.vector_store import VectorRegister, VectorStore
from cel.rag.text2vec.utils import Embedding, Text2VectorProvider



@dataclass
class VectorRegisterResult(VectorRegister):
    """Vector register for ChromaDB"""
    distance: float
    

class ChromaStore(VectorStore):
    
    def __init__(self, text2vec_provider: Text2VectorProvider, collection_name: str = "my_collection"):
        log.debug(f"Instantiate ChromaStore with collection_name: {collection_name}")
        self.text2vec = text2vec_provider
        self.client = chromadb.Client()
        self.collection_name = collection_name
        
        # create collection if not exists
        collections = self.client.list_collections()
        for collection in collections:
            if collection.name == collection_name:
                self.collection = collection
                log.debug(f"Collection {collection_name} already exists")
                break
        else:
            log.debug(f"Creating collection {collection_name}")
            self.collection = self.client.create_collection(
                            name=collection_name,
                        )

    def get_vector(self, id: str) -> VectorRegister:
        """Get a vector chromadb by id"""
        v = self.collection.get(id, include=['documents', 'metadatas', 'embeddings'])
        # create a VectorRegister object
        return VectorRegister(id=id,
                                vector=v['embeddings'][0],
                                text=v['documents'][0],
                                metadata=v['metadatas'][0])
        

    def get_similar(self, vector: Embedding, top_k) -> list[VectorRegisterResult]:
        """Get the most similar vectors to the given vector"""
        res = self.collection.query(
            query_embeddings=[vector],
            n_results=top_k,
            include=['documents', 'metadatas', 'distances', 'embeddings']
        )
        
        return [VectorRegisterResult(id=id, 
                               vector=embedding, 
                               distance=distance,
                               text=text,
                               metadata=metadata) for id, embedding, distance, text, metadata in zip(res['ids'][0],
                                                                                        res['embeddings'][0],
                                                                                        res['distances'][0],
                                                                                        res['documents'][0], 
                                                                                        res['metadatas'][0])]   
                

    def search(self, query: str, top_k: int = 1) -> list[VectorRegisterResult]:
        """Search for vectors by a query"""
        
        vector = self.text2vec.text2vec(query)
        res = self.collection.query(
            query_embeddings=[vector],
            n_results=top_k,
            include=['documents', 'metadatas', 'distances', 'embeddings']
        )

        # convert QueryResult to VectorRegister
        # res sample: {'data': None, 'distances': [[0.7714858651161194, 1.6371256113052368]], 'documents': [['This is a document about parrots', 'This is a document about pineapple']], 'ids': [['4', '0']], 'metadatas': [[{'metadata': 'metadata'}, {'metadata': 'metadata'}]]}
        return [VectorRegisterResult(id=id, 
                               vector=embedding, 
                               distance=distance,
                               text=text,
                               metadata=metadata) for id, embedding, distance, text, metadata in zip(res['ids'][0],
                                                                                        res['embeddings'][0],
                                                                                        res['distances'][0],
                                                                                        res['documents'][0], 
                                                                                        res['metadatas'][0])]   
        

        
        

    def upsert(self, id: str, vector: Embedding, text: str, metadata: dict):
        """Upsert a vector to the store"""
        self.collection.add(
            documents=text,
            ids=[id],
            metadatas=metadata
        )
        
    def upsert_text(self, id: str, text: str, metadata: dict):
        """Upsert a vector to the store"""
        try:
            vector =  self.text2vec.text2vec(text)
            ## if vector is a string '[n,n,n.....]', convert it to a list
            if isinstance(vector, str):
                vector = [float(i) for i in vector[1:-1].split(',')]
            ## if vector is a list of strings ['n','n','n'.....], convert it to a list of floats
            if isinstance(vector[0], str):
                vector = [float(i) for i in vector] 
        except Exception as e:
            log.error(f"Error converting vector: {e}")
            vector = None
        
        
        self.collection.add(
            ids=[id],
            documents=text,
            metadatas=metadata,
            embeddings=vector
        )

    def delete(self, id):
        """Delete a vector from the store"""
        self.collection.delete(id)


from dataclasses import dataclass
from typing import Callable


class RequestMode:
    STREAM = "stream"
    SINGLE = "single"


@dataclass
class FunctionResponse:
    text: str

    """ the execution function response will be sent as a stream of chunks or as a single response """
    request_mode: str = RequestMode.SINGLE

    """ If you want to disable direct response from Asiistant AI, use callback to handle the response.
    It is useful when you want to handle the response in a different way, like sending the response to a different service.
    Another use case is when oyu need to use specific platform features to send the response, like interactives messages, buttons, etc.
    """
    callback: Callable = None
    
    


from abc import ABC

class MessageAttachmentType:
    IMAGE = "image"
    AUDIO = "audio"
    VOICE = "voice"
    VIDEO = "video"
    DOCUMENT = "document"
    LINK = "link"
    LOCATION = "location"
    CONTACT = "contact"
    CUSTOM = "custom"

class MessageAttachment(ABC):
    def __init__(self, type: MessageAttachmentType = None) -> None:
        self.type = type
        
class ContactAttachment(MessageAttachment):
    def __init__(self, 
                 name: str = None, 
                 metadata: any = None):
        super().__init__(type=MessageAttachmentType.CONTACT)
        self.name: str = name
        self.metadata: any = metadata
        
    def to_dict(self):
        data = {
            'name': self.name,
            'metadata': self.metadata
        }
        return data

    @classmethod
    def from_dict(cls, attachment_dict):
        return ContactAttachment(
            name=attachment_dict.get("name"),
            metadata=attachment_dict.get("metadata")
        )

    def __str__(self):
        return f"ContactAttachment: {self.name}"

class LocationAttachment(MessageAttachment):
    def __init__(self, 
                 latitude: float, 
                 longitude: float, 
                 description: str = None,
                 metadata: any = None):
        super().__init__(type=MessageAttachmentType.LOCATION)
        self.latitude: float = latitude
        self.longitude: float = longitude
        self.metadata: any = metadata
        self.description: str = description

    def to_dict(self):
        data = {
            'latitude': self.latitude,
            'longitude': self.longitude,
            'metadata': self.metadata
        }
        return data

    @classmethod
    def from_dict(cls, attachment_dict):
        return LocationAttachment(
            latitude=attachment_dict.get("latitude"),
            longitude=attachment_dict.get("longitude"),
            metadata=attachment_dict.get("metadata")
        )

    def __str__(self):
        return f"LocationAttachment: {self.latitude}, {self.longitude}"
    
    def __repr__(self):
        return f"LocationAttachment: {self.latitude}, {self.longitude}"

class FileAttachment(MessageAttachment):
    def __init__(self, 
                 title: str = None, 
                 description: str = None, 
                 content: any = None, 
                 mimeType: str = None,
                 file_url: str = None,
                 metadata: any = None,
                 type: MessageAttachmentType = None):
        self.title: str = title
        self.description: str = description
        self.content: any = content
        self.mimeType: str = mimeType
        self.file_url: str = file_url
        self.metadata: any = metadata
        self.type = type

    def to_dict(self):
        data = {
            'title': self.title,
            'description': self.description,
            'content': self.content,
            'mimeType': self.mimeType,
            'file_url': self.file_url,
            'metadata': self.metadata
        }
        return data

    @classmethod
    def from_dict(cls, attachment_dict):
        return FileAttachment(
            title=attachment_dict.get("title"),
            description=attachment_dict.get("description"),
            content=attachment_dict.get("content"),
            mimeType=attachment_dict.get("mimeType"),
            metadata=attachment_dict.get("metadata"),
            file_url=attachment_dict.get("file_url")
        )

    def __str__(self):
        return f"MessageAttachment: {self.title}"
    
    def __repr__(self):
        return f"MessageAttachment: {self.title}"
 


import re


def format_to_e164(phone_number):
    # Eliminar todos los caracteres que no sean dígitos
    digits_only = re.sub(r'\D', '', phone_number)
    
    # Agregar el signo '+' al inicio
    return f'+{digits_only}'




# from .attachment import MessageAttachmentType, MessageAttachment, LocationAttachment, FileAttachment
# from .conversation_lead import ConversationLead
# from .conversation_peer import ConversationPeer
# from .message_gateway_context import MessageGatewayContext
# from .message import Message
# from .base_connector import BaseConnector


from .deepgram_stt import DeepgramSTTMiddleware
from .geodecoding import GeodecodingMiddleware
from .in_mem_blacklist import InMemBlackListMiddleware
from .session_middleware import SessionMiddleware
from .redis_blacklist import RedisBlackListMiddleware
from .redis_blacklist_async import RedisBlackListAsyncMiddleware


class RunMode:
    """
    A class to represent the different modes of running a Telegram connector.

    Attributes:
    -----------
    WEBHOOK : str
        A constant representing the webhook mode.
    POLLING : str
        A constant representing the polling mode.

    Methods:
    --------
    get_modes():
        Returns a list of all available run modes.
    is_valid(mode: str):
        Checks if the provided mode is a valid run mode.
    get_mode(mode: str):
        Returns the mode if it is valid, otherwise raises an exception.
    get_default():
        Returns the default run mode.
    """

    WEBHOOK = "webhook"
    POLLING = "polling"

    @staticmethod
    def get_modes():
        """
        Returns a list of all available run modes.

        Returns:
            list: A list containing all available run modes.
        """
        return [RunMode.WEBHOOK, RunMode.POLLING]

    @staticmethod
    def is_valid(mode: str):
        """
        Checks if the provided mode is a valid run mode.

        Parameters:
            mode (str): The mode to be checked.

        Returns:
            bool: True if the mode is valid, False otherwise.
        """
        return mode in RunMode.get_modes()

    @staticmethod
    def get_mode(mode: str):
        """
        Returns the mode if it is valid, otherwise raises an exception.

        Parameters:
            mode (str): The mode to be returned.

        Returns:
            str: The valid mode.

        Raises:
            Exception: If the mode is not valid.
        """
        if RunMode.is_valid(mode):
            return mode
        else:
            raise Exception(f"Invalid run mode: {mode}. Valid modes are: {RunMode.get_modes()}")

    @staticmethod
    def get_default():
        """
        Returns the default run mode.

        Returns:
            str: The default run mode.
        """
        return RunMode.WEBHOOK


from cel.stores.state.base_state_provider import BaseChatStateProvider
from loguru import logger as log

class InMemoryStateProvider(BaseChatStateProvider):

    def __init__(self, key_prefix: str = "s"):
        super().__init__()
        log.warning(f"Create InMemoryStateProvider - Avoid using this in production.")
        self.store = {}
        self.prefix = key_prefix

    def get_key(self, sessionId):
        return f"{self.prefix}:{sessionId}"

    async def set_key_value(self, sessionId: str, key: str, value, ttl_in_seconds=None):
        hash_key = self.get_key(sessionId)
        if hash_key not in self.store:
            self.store[hash_key] = {}
        self.store[hash_key][key] = value

    async def get_key_value(self, sessionId: str, key: str):
        hash_key = self.get_key(sessionId)
        if hash_key in self.store and key in self.store[hash_key]:
            return self.store[hash_key][key]
        return None

    async def clear_store(self, sessionId: str):
        hash_key = self.get_key(sessionId)
        if hash_key in self.store:
            del self.store[hash_key]

    async def clear_all_stores(self):
        self.store = {}

    async def get_store(self, sessionId: str):
        hash_key = self.get_key(sessionId)
        if hash_key in self.store:
            return self.store[hash_key]
        return None

    async def set_store(self, sessionId: str, store, ttl=None):
        hash_key = self.get_key(sessionId)
        self.store[hash_key] = store


from cel.gateway.model.base_connector import BaseConnector
from cel.connectors.telegram.model.telegram_attachment import TelegramAttachment
from cel.connectors.telegram.model.telegram_lead import TelegramLead
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.message import Message


class TelegramMessage(Message):
    
    def __init__(self, 
                 lead: ConversationLead, 
                 text: str = None, 
                 metadata: dict = None, 
                 date: int = None,
                 attachments: list[TelegramAttachment] = None
                ):
        super().__init__(lead, text=text, date=date, metadata=metadata, attachments=attachments)
    
    
    def is_voice_message(self):
        #  check if the message has a voice attachment
        if self.attachments:
            for attach in self.attachments:
                if attach.type == "voice":
                    return True
        return False
    
    @classmethod
    async def load_from_message(cls, message_dict, token: str, connector: BaseConnector = None):
        msg = message_dict.get("message")
        # get text from message or caption if it is a media message
        text = msg.get("text") or msg.get("caption")
        
        # if text begins with /start, it is a command
        if text and text.startswith("/start"):
            # decode the arguments
            args = text.split(" ")[1:]
            # decode string from base64
            import base64
            text = base64.b64decode(args[0]).decode("utf-8")
        
        date = msg.get("date")
        metadata = {'raw': msg}
        lead = TelegramLead.from_telegram_message(msg, connector=connector)
        attach = await TelegramAttachment.load_from_message(message_dict, token)
        return TelegramMessage(lead=lead, text=text, date=date, metadata=metadata, attachments=[attach] if attach else None)
        

    def __str__(self):
        return f"TelegramMessage: {self.text}"
        
    def __repr__(self):
        return f"TelegramMessage: {self.text}"
    
    


from loguru import logger as log
from cel.gateway.model.message import Message
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.base_connector import BaseConnector
from cel.connectors.livekit.model.livekit_lead import LiveKitLead

class LiveKitMessage(Message):
    """
    LiveKitMessage class to represent a message in the LiveKit connector.
    """

    def __init__(self, lead: ConversationLead, text: str = None):
        """
        Initialize the LiveKitMessage instance.

        Args:
            lead (ConversationLead): The lead associated with the message.
            text (str): The text of the message.
        """
        super().__init__(lead, text=text)

    def is_voice_message(self):
        return False
    
    @classmethod
    async def load_from_message(cls, request, connector: BaseConnector = None):
        """
        Load a LiveKitMessage from a message dictionary.

        Args:
            message_dict (dict): The message dictionary.
            token (str): The token for authentication.
            connector: The connector instance.

        Returns:
            LiveKitMessage: The loaded LiveKitMessage instance.
        """
        log.debug(f"[LiveKit] Raw request in load_from_message: {request}")
        user_message = request.get("user_text")
        
        if not user_message:
            raise ValueError("No message found in the message_dict")

        lead = LiveKitLead.from_message(request, connector=connector)
        return LiveKitMessage(lead=lead, text=user_message)
    
    def __str__(self):
        return f"LiveKitMessage: {self.text}"
    
    def __repr__(self):
        return f"LiveKitMessage: {self.text}"



"""
LiveKitAdapter masquerades as a livekit.LLM and communicates with Celai
through HTTP API calls with proper handling of stream formatting.
"""

from typing import Optional, Any, List
from loguru import logger as log

# If livekit-agents not installed raise ImportError
try:
    from livekit.agents import llm
    from livekit.agents import llm, FunctionTool  
    from livekit.agents.tts import SynthesizeStream
    from livekit.agents.types import APIConnectOptions, DEFAULT_API_CONNECT_OPTIONS
    from livekit.agents.utils import shortuuid
except ImportError:
    raise ImportError(
        "LiveKit agents library not installed. Please install it using 'pip install livekit-agents'."
    )

try:
    import httpx
except ImportError:
    raise ImportError(
        "httpx library not installed. Please install it using 'pip install httpx'."
    )


class FlushSentinel(str, SynthesizeStream._FlushSentinel):
    """
    A sentinel object that indicates when to flush the stream.
    """
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)


class LiveKitStream(llm.LLMStream):
    def __init__(
        self,
        llm: llm.LLM,
        *,
        chat_ctx: llm.ChatContext,
        session_id: str,
        api_url: str,
        timeout: float = 30.0,
        tools: Optional[List[FunctionTool]] = None,
        conn_options: APIConnectOptions | None = None,
        **kwargs,
    ):
        super().__init__(llm, chat_ctx=chat_ctx, tools=tools, conn_options=conn_options)
        self._session_id = session_id
        self._api_url = api_url
        self._timeout = timeout
        self._kwargs = kwargs

    @staticmethod
    def _create_livekit_chunk(
        content: str,
        *,
        id: str | None = None,
    ) -> llm.ChatChunk:
        """Generates a ChatChunk with the new schema (id + delta)."""
        return llm.ChatChunk(
            id=id or shortuuid(),
            delta=llm.ChoiceDelta(role="assistant", content=content),
        )
    
    def _to_message(self, msg: llm.ChatMessage) -> str:
        """
        Convert a LiveKit ChatMessage to text.
        
        Args:
            msg: The LiveKit chat message
            
        Returns:
            The extracted text from the message
        """
        if isinstance(msg.content, str):
            return msg.content
        elif isinstance(msg.content, list):
            text_parts = []
            for c in msg.content:
                if isinstance(c, str):
                    text_parts.append(c)
                elif isinstance(c, dict) and c.get("type") == "text":
                    text_parts.append(c.get("text", ""))
                elif hasattr(c, "text"):
                    text_parts.append(c.text)
            return " ".join(text_parts)
        return ""
    
    async def _to_livekit_chunk(self, content: str | Any) -> llm.ChatChunk | None:
        if not content:
            return None
        chunk_id = getattr(content, "id", None)
        if isinstance(content, str):
            return self._create_livekit_chunk(content, id=chunk_id)
        if isinstance(content, dict) and "content" in content:
            return self._create_livekit_chunk(content["content"], id=content.get("id", chunk_id))
        return None
    
    async def _run(self):
        """Process the incoming message and stream the response"""
        try:
            # Find the most recent user message
            input_text = None
            for m in reversed(self.chat_ctx.items):
                if m.role == "user":
                    log.debug(f"User message found: {m}")
                    input_text = self._to_message(m)
                    break
            
            if not input_text:
                log.warning("No user message found in chat context")
                return
            
            log.debug(f"Processing via API for session {self._session_id}")
            
            # Prepare the request payload
            payload = {
                "session_id": self._session_id,
                "user_text": input_text,
                **self._kwargs
            }
            
            # Make the request and stream the response
            async with httpx.AsyncClient() as client:
                async with client.stream(
                    "POST",
                    self._api_url,
                    json=payload,
                    timeout=self._timeout,
                ) as response:
                    # Check for successful response
                    if response.status_code != 200:
                        err_txt = await response.text()
                        raise RuntimeError(f"API error => status {response.status_code}: {err_txt}")

                    
                    # Process the response line by line
                    async for raw in response.aiter_lines():
                        if not raw.startswith("data: "):
                            continue

                        content = raw.removeprefix("data: ").strip("\r\n")

                        if not content:
                            continue

                        # -- flush sentinel —
                        if content == "<FLUSH>":
                            self._event_ch.send_nowait(
                                self._create_livekit_chunk(FlushSentinel())
                            )
                            continue

                        # -- token normal: lo enviamos tal cual, sin tocar espacios —
                        self._event_ch.send_nowait(
                            self._create_livekit_chunk(content)
                        )
                    
                    # End of stream
                    self._event_ch.send_nowait(
                        self._create_livekit_chunk(FlushSentinel())
                    )
                
        except Exception as e:
            log.error(f"Error in LiveKitStream._run: {e}")
            error_message = f"Error processing your request: {str(e)}"
            if chunk := await self._to_livekit_chunk(error_message):
                self._event_ch.send_nowait(chunk)
            
            # Always send a flush sentinel on error to ensure the stream completes
            self._event_ch.send_nowait(
                self._create_livekit_chunk(FlushSentinel())
            )


class LiveKitAdapter(llm.LLM):
    """
    Adapter class that implements LiveKit's LLM interface
    and delegates to Celai API for actual processing.
    """
    
    def __init__(
        self, 
        api_url: str,
        timeout: float = 30.0,
        **kwargs
    ):
        """
        Initialize the adapter with parameters for API communication.
        
        Args:
            api_url: URL for the Celai API endpoint
            timeout: Timeout for HTTP requests in seconds
            **kwargs: Additional parameters to pass to the API
        """
        super().__init__()
        
        # Store configuration
        self._api_url = api_url
        self._timeout = timeout
        self._kwargs = kwargs
        
        log.info(f"Initialized LiveKitAdapter using API URL: {api_url}")
    
    def chat(
        self,
        *,
        chat_ctx: llm.ChatContext,
        tools: list[FunctionTool] | None = None,
        conn_options: APIConnectOptions = DEFAULT_API_CONNECT_OPTIONS,
        **kwargs: Any,
        
    ) -> llm.LLMStream:
        """
        Create a stream for LiveKit chat.
        
        This method is called by LiveKit to process a message and get a response.
        
        Args:
            chat_ctx: The chat context from LiveKit
            fnc_ctx: The function context from LiveKit
            conn_options: Connection options from LiveKit
            
        Returns:
            A LiveKit stream for handling the chat
        """
        base_id = chat_ctx.items[0].id if chat_ctx.items else shortuuid()
        session_id = f"livekit:{base_id}"
        return LiveKitStream(
            llm=self,
            chat_ctx=chat_ctx,
            session_id=session_id,
            api_url=self._api_url,
            timeout=self._timeout,
            tools=tools,
            conn_options=conn_options,
            **self._kwargs,
        )



from typing import Union
import cachetools
from .key_value_store import KeyValueStore, ListStore


class MemoryCache(KeyValueStore):
    """MemoryCache is a CacheStore implementation that stores cache data in memory using cachetools.LRUCache.
    
    Args:
        key_prefix (str): A key prefix for the cache store.
        memory_maxsize (int): The maximum number of items to store in the cache.
    """
    
    
    def __init__(self, key_prefix, memory_maxsize=1000):
        self.cache = cachetools.LRUCache(maxsize=memory_maxsize)
        self.key_prefix = key_prefix

    def get(self, key, callback=None):
        """Retrieves the value associated with the given key from the cache.
        
        If the key is not in the cache, it calls the provided callback function to compute the value.
        
        Args:
            key (str): The cache key.
            callback (callable): A function that computes the value for the given key if it is not in the cache (optional).
        
        Returns:
            The value associated with the key, or None if the key is not in the cache.
        """
        
        data = self.cache.get(key)
        if data is not None:
            return data
        
        if callback:
            data = callback()
            if data is not None:
                self.set(key, data)
                return data

        return None

    
    def set(self, key, value):
        """Sets the value for the given key in the cache.
        
        Args:
            key (str): The cache key.
            value (any): The value to store in the cache.
        """
        self.cache[key] = value
        
    def get_all(self):
        """Returns all items in the cache.
        
        Returns:
            A list of key-value pairs in the cache.
        """
        return self.cache.items()

    def delete(self, key):
        """Deletes the key from the cache.
        
        Args:
            key (str): The cache key.
        """
        self.cache.pop(key, None)
       
    def clear(self):
        """Clears the cache."""
        self.cache.clear()
    
    def all(self):
        """Returns all items in the cache.
        
        Returns:
            A list of key-value pairs in the cache.
        """
        return self.cache.items()
    



import asyncio
import json
from typing import Any, Dict
from urllib.parse import urljoin
from fastapi import HTTPException
import shortuuid
from cel.assistants.base_assistant import BaseAssistant
from cel.gateway.message_gateway import StreamMode
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.message import Message
from cel.gateway.model.message_gateway_context import MessageGatewayContext
from cel.gateway.model.middleware import BaseMiddleware
from cel.gateway.model.outgoing.outgoing_message import OutgoingMessage
from cel.middlewares.chatwoot.conversation_manager import ConversationManager
from cel.middlewares.chatwoot.model import ContactLead
from loguru import logger as log


class ChatwootMiddleware(BaseMiddleware):

    def __init__(self, base_url: str, access_key: str, account_id: str, inbox_name: str, auto_create_inbox: bool = True):
        self.conversation_manager = None
        self.base_url = base_url    
        self.access_key = access_key
        self.account_id = account_id
        self.inbox_name = inbox_name
        self.auto_create_inbox = auto_create_inbox
        # random generate a code for the webhook using short uuid
        self.security_token = shortuuid.uuid()
        
    def create_contact_from_incoming_message(self, message: Message) -> ContactLead:
        
        return ContactLead(
            identifier=message.lead.get_session_id(),
            name=message.lead.conversation_from.name,
            phone_number=message.lead.conversation_from.phone,
            celai_lead=ConversationLead.serialize(message.lead)
        )

            
    async def startup(self, ctx: MessageGatewayContext):
        
        endpoint_url = await self.setup_routes(ctx.app)
        
        webhook_url = urljoin(ctx.webhook_url, endpoint_url)
        
        self.conversation_manager = ConversationManager(
            base_url=self.base_url,
            access_key=self.access_key,
            webhook_url=webhook_url,
            account_id=self.account_id,
            inbox_name=self.inbox_name
        )
    
        await self.conversation_manager.init(auto_create_inbox=self.auto_create_inbox)
        
        
    async def setup_routes(self, app):
        log.debug("Setting up Chatwoot middleware routes")
        prefix = "/middleware/chatwoot"
        
        
        # Add echo route
        @app.get(prefix + "/echo")
        async def echo():
            return {"status": "ok"}
        
        @app.post(prefix + "/messages/{security_token}", 
                 summary="Incoming message",
                 description="This endpoint receives incoming messages from Chatwoot")
        async def incoming_message(security_token: str, payload: Dict[Any, Any]):
            # check code is == self.code
            if security_token != self.security_token:
                raise HTTPException(status_code=401, detail="Unauthorized")
            
            # message_type = payload.get("message_type")
            try:
                log.debug(f"Incoming message: {payload}")
                
                if payload.get("event") == "conversation_updated":
                    # A conversation was updated
                    # we need to check if conversation was assigned to an agent
                    # if the conversation was assigned to an agent, we need to mark the conversation
                    # so Celai does not respond to it
                    changes = payload.get("changed_attributes", [])
                    # if changes array contains an object with assignee_id property defined
                    assignee_id = next((item for item in payload.get("changed_attributes", []) if "assignee_id" in item),{}).get('assignee_id')
                    if assignee_id:
                        # This means the conversation was assigned to an agent
                        # we need to mark the conversation as assigned in Celai
                        # so Celai does not respond to it
                    
                        # Conversation was assigned to an agent
                        messages = payload.get("messages", [])
                        if messages:
                            # Get the last message
                            last_message = messages[-1]
                            conversation_id = last_message.get("conversation_id")
                            if conversation_id:
                                # Mark conversation as assigned
                                conversation = self.conversation_manager.get_conversation_by_id(conversation_id)
                                if conversation:
                                    previous_value = assignee_id.get("previous_value")
                                    current_value = assignee_id.get("current_value")
                                    if current_value:
                                        conversation.assigned = True
                                        log.debug(f"Conversation {conversation_id} assigned to agent, marking as assigned in Celai")
                                    elif previous_value and not current_value:
                                        conversation.assigned = False
                                        log.debug(f"Conversation {conversation_id} unassigned from agent, marking as unassigned in Celai")
                                    
                
                if payload.get("message_type") == "outgoing" and not payload.get("private"):
                    # This is an outgoing message from Chatwoot
                    # we must send it to the client
                    content = payload.get("content", "")
                    if content.startswith("[LOLA]: "):
                        # This is a message from Celai, we can ignore it
                        log.debug("Ignoring Celai message from Chatwoot")
                        return {"status": "ok"}
                    
                    celai_lead = payload.get("conversation", {}).get("custom_attributes", {}).get("celai_lead")
                    if not celai_lead:
                        log.error("No celai_lead found in the message payload", payload=payload)
                        return {"status": "ok"}
                    
                    # Deserialize the lead, context aware
                    lead = ConversationLead.deserialize(celai_lead)
                    content = payload.get("content", "")
                    
                    await lead.connector.send_text_message(lead=lead, text=content)
                    log.debug(f"Message from Chatwoot sent to lead: {lead.get_session_id()} with content: {content}")
                    
            except Exception as e:
                log.error(f"Chatwoot Middleware: Error processing incoming message: {e}")
            
            finally:
                return {"status": "ok"}
            

        message_endpoint_url = f"{prefix}/messages/{self.security_token}"
        return message_endpoint_url
    
    
    async def incoming_message(self,  message: Message, connector: BaseConnector, assistant: BaseAssistant):
        
        try:
            assert self.conversation_manager is not None, "ChatwootMiddleware is not initialized"
            
            cto = self.create_contact_from_incoming_message(message)
            # await self.conversation_manager.send_incoming_text_message(cto, message.text)
            asyncio.create_task(self.conversation_manager.send_incoming_text_message(cto, message.text))
            
            
            conv = await self.conversation_manager.get_conversation(cto.identifier)
            if conv and conv.assigned:
                log.warning(f"Conversation {conv.id} is assigned to an agent, skipping Celai response")
                return False
            
            
            return True
        except Exception as e:
            log.error(f"Middleware Chatwoot: Error processing incoming message: {e}")
            return True
    
    
    
    async def outgoing_message(self, 
                               message: OutgoingMessage, 
                               connector: BaseConnector, 
                               assistant: BaseAssistant,
                               is_partial: bool = False, 
                               is_summary: bool = False,
                               mode: StreamMode = None):
        
        try:
            assert self.conversation_manager is not None, "ChatwootMiddleware is not initialized"
                    
            if (mode == StreamMode.FULL and not is_summary) or (mode != StreamMode.FULL and is_summary):
                cto = self.create_contact_from_incoming_message(message)
                asyncio.create_task(self.conversation_manager.send_outgoing_text_message(cto,  '[LOLA]: ' + str(message)))
            
            return True
        except Exception as e:
            log.error(f"Middleware Chatwoot: Error processing outgoing message: {e}")
            return True


from abc import ABC
import inspect
import re
import time
from typing import Any, Dict, Optional
from fastapi import BackgroundTasks, HTTPException, Request
from fastapi.responses import RedirectResponse
from pydantic import BaseModel
import json
from dataclasses import asdict, dataclass
from cel.assistants.base_assistant import BaseAssistant
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.message import Message
from loguru import logger as log

from cel.gateway.model.message_gateway_context import MessageGatewayContext
from .utils import decode_jwt, generate_encryption_key, generate_link, generate_secret_key

@dataclass  
class CallbackEntry(ABC):
    handler: callable
    redirect_url: str = None,
    single_use: bool = True

    

class HttpCallbackProvider(ABC):

    def __init__(self, 
                 endpoint: str = "callback",
                 http_verb: str = "GET",
                 secret_key: str = None,
                 encryption_key: str = None,
                ):
        
        self.endpoint = endpoint
        self.external_url = None
        self.secret_key = secret_key or generate_secret_key()
        self.encryption_key = encryption_key or generate_encryption_key()

        if http_verb.upper() not in ["GET", "POST"]:
            raise ValueError("HTTP Verb must be GET or POST")
        
        self.http_verb = http_verb
        
        self.handlers = {}



    async def _handle_callback(self, 
                                token: str,  
                                request: Request):
        
        if not token:
            raise HTTPException(status_code=401, detail="Token not found")
        
        try:
        #  Decode token
            json_data = decode_jwt(token, self.secret_key, self.encryption_key)
            json_lead = json_data.get("lead")
            lead = ConversationLead.deserialize(json_lead)
            log.debug(f"Decoded lead: {lead}")
            
            # await self.gateway.assistant.call_event('callback', lead, None, lead.connector, data=request.query_params)
            handler_id = json_data.get("handler_id")
            entry = self.handlers.get(handler_id)
            
            if not entry or not isinstance(entry, CallbackEntry):
                raise HTTPException(status_code=401, detail="Handler not found")
            
            handler = entry.handler
            params_dict = {}
            try: 
                # Convert QueryParams to dict
                params_dict.update(dict(request.query_params))
                # get request body if it's a POST request
                if self.http_verb.upper() == "POST":
                    body = await request.body()
                    try:
                        body_dict = json.loads(body)
                        params_dict.update(body_dict)
                    except Exception as e:
                        log.error(f"Error parsing body: {e}")
                        raise HTTPException(status_code=401, detail="Error parsing body")
                    
                # handler is coroutine function?
                res = None
                if inspect.iscoroutinefunction(handler):
                    res = await handler(lead, params_dict)
                else:
                    res = handler(lead, params_dict)
                    
                if res: 
                    log.debug(f"Handler response: {res}")
                    return res
                    
            except Exception as e:
                log.error(f"Error calling handler with lead: {lead}: {e}")
                raise HTTPException(status_code=401, detail="Error calling handler")

            if entry.single_use:
                self.handlers.pop(handler_id)
            
            if entry.redirect_url:
                log.debug(f"Redirecting to lead: {lead} to {entry.redirect_url}")
                return RedirectResponse(url=entry.redirect_url)
            
            
            return {
                "status": "ok"
            }
        
        except Exception as e:
            log.error(f"Error processing callback: {e}")
            raise HTTPException(status_code=401, detail="Error processing callback")

    

    def setup(self, ctx: MessageGatewayContext):
        log.debug("Setting up Http Callback Manager")
        self.external_url = ctx.webhook_url
        app = ctx.app

        if self.http_verb.upper() == "GET":
            app.get("/" + self.endpoint + "/{token}")(self._handle_callback)
            
        if self.http_verb.upper() == "POST":
            app.post("/" + self.endpoint + "/{token}")(self._handle_callback)



    def create_callback(self, lead: ConversationLead, handler: callable, ttl: int = 180, redirect_url: str = None, single_use: bool = True):
        """ Create a callback link for a lead """
        
        assert isinstance(lead, ConversationLead), "Lead must be a ConversationLead object"
        assert self.external_url, "External URL must be set"
        
        callback_id = id(handler)
        data = {
            "lead": ConversationLead.serialize(lead),
            "handler_id": callback_id
        }
    
        link = generate_link(
            base_url=self.external_url + "/" + self.endpoint,
            secret_key=self.secret_key,
            encryption_key=self.encryption_key,
            data=data,
            ttl=ttl
        )
        
        self.handlers[callback_id] = CallbackEntry(
            redirect_url=redirect_url,
            handler=handler,
            single_use=single_use
        )
        
        log.debug(f"Generated link: {link}")
        return link



from abc import ABC, abstractmethod
from typing import Any

class BaseVoiceProvider(ABC):
    
    @abstractmethod
    def TTS(self, text: str, voice: str = None, settings: Any = None):
        raise NotImplementedError
    
    @abstractmethod
    def STT(self, audio: bytes | str) -> str:
        raise NotImplementedError


"""
Unofficial Python wrapper for the WhatsApp Cloud API.
"""
from __future__ import annotations
import json
from typing import Any, BinaryIO, Callable, Dict
import aiohttp
from fastapi import APIRouter, BackgroundTasks, Request
from loguru import logger as log
import aiohttp
import shortuuid
from cel.connectors.telegram.telegram_connector import hash_token
from cel.connectors.whatsapp.components.list_item import ListItem
from cel.gateway.model.base_connector import BaseConnector
from cel.connectors.whatsapp.components.reply_button import ReplyButton
from cel.connectors.whatsapp.constants import BASE_URL
from cel.connectors.whatsapp import WhatsappLead
from cel.connectors.whatsapp.model.whatsapp_message import WhatsappMessage
from cel.connectors.whatsapp.utils import build_headers
from cel.gateway.message_gateway import StreamMode
from cel.gateway.model.message_gateway_context import MessageGatewayContext
from cel.gateway.model.outgoing import OutgoingMessage,\
                                            OutgoingMessageType,\
                                            OutgoingLinkMessage,\
                                            OutgoingSelectMessage,\
                                            OutgoingTextMessage
from cel.gateway.model.outgoing.outgoing_message_buttons import OutgoingButtonsMessage
from .functions.utils import changed_field, is_message, is_reaction




async def nothing(*args):
    pass


class WhatsappConnector(BaseConnector):
    def __init__(self, 
                 token: str = None,
                 phone_number_id: str = None,
                 allowed_numbers: list[str] = None,
                 endpoint_prefix: str = None,
                 stream_mode: StreamMode = StreamMode.SENTENCE,
                 verify_token: str = None,
                 ssl: bool = False):
        """
        Initialize the Async WhatsApp Cloud API Connector with the Meta Access Token and the Phone Number Id

        Args:
            - token[str]: The Meta Access Token
            - phone_number_id[str]: The Meta Phone Number Id
            - verify_token[str]: The verification token, defaults to a random string
            - webhook_prefix[str]: The prefix for the webhook, defaults to "/whatsapp"
            - stream_mode[StreamMode]: The stream mode for the gateway, defaults to StreamMode.SENTENCE
        """

        # Verify the token and phone number id
        assert token is not None, "Token not provided"
        assert phone_number_id is not None, "Phone number id not provided"
        # assert verify_token is not None, "Verify token not provided"
        
        self.token = token
        self.phone_number_id = phone_number_id
        self.allowed_numbers = allowed_numbers
        self.base_url = BASE_URL
        self.url = f"{self.base_url}/{phone_number_id}/messages"
        self.verify_token = verify_token or shortuuid.uuid()
        self.endpoint_prefix = endpoint_prefix or "/whatsapp"
        self.stream_mode = stream_mode
        self.verification_handler = nothing
        self.ssl = ssl
        log.debug("Whatsapp Connector initialized")
        
        
        
    def __build_routes(self, endpoint_prefix: str):
        assert endpoint_prefix is not None, "Endpoint prefix not provided"
        
        def shutdown():
            log.warning("Shutting down whatsapp connector")
        
        def startup():
            log.debug(f"Starting up whatsapp connector")

        router = APIRouter(prefix=endpoint_prefix,
                            on_startup=[startup],
                            on_shutdown=[shutdown])

        # Define the routes
        # ---------------------------------------------------------------
        @router.get("/")
        async def verify_endpoint(r: Request):
            if r.query_params.get("hub.verify_token") == self.verify_token:
                log.info("Verified webhook")
                challenge = r.query_params.get("hub.challenge")
                await self.verification_handler(challenge)
                return int(challenge)
            log.error("Webhook Verification failed")
            await self.verification_handler(False)
            return {"success": False}

        def get_display_phone_number(data: dict) -> str:
            """Safely retrieve the display phone number from the incoming data."""
            try:
                return data["entry"][0]["changes"][0]["value"]["metadata"]["display_phone_number"]
            except (IndexError, KeyError, TypeError) as e:
                log.error(f"Error retrieving display phone number: {e}")
                return None

        @router.post("/")
        async def hook(r: Request, background_tasks: BackgroundTasks):
            try:
                # Handle Webhook Subscriptions
                log.info("Received message from Whatsapp")
                data = await r.json()
                log.debug(json.dumps(data, indent=2))
                display_phone_number = get_display_phone_number(data)
                if self.allowed_numbers and display_phone_number not in self.allowed_numbers:
                    log.warning(f"Display phone number {display_phone_number} is not allowed.")
                    return {"success": True}
                else:
                    background_tasks.add_task(self.__process_message, data)
                    return {"success": True}
            except Exception as e:
                # Avoid replaying the message
                log.error(f"Error processing message: {e}")
                return {"success": True}
        # ---------------------------------------------------------------
        return router
    
    async def __process_message(self, data: dict):      
        changed_field = self.changed_field(data)
        log.debug(f"Changed field: {changed_field}")
        if changed_field == "messages":
            if self.is_reaction(data):
                log.debug("Reaction received")
                return
            new_message = self.is_message(data)
            log.debug(f"New message: {new_message}")
            if new_message:
                msg = await WhatsappMessage.load_from_message(data=data, 
                                                        token=self.token, 
                                                        phone_number_id=self.phone_number_id, 
                                                        connector=self)
                await self.__mark_as_read(msg.id)
                if self.gateway:
                    async for m in self.gateway.process_message(msg, mode=self.stream_mode):
                        pass

    def on_verification(self, handler: callable):
        """
        Set the handler for verification
        Args:
            handler[function]: The handler function
        """
        assert callable(handler), "Handler must be a function"
        self.verification_handler = handler

    def name(self) -> str:
        """ This method returns the name of the connector. This is used to identify the connector
        In order to allow multiple connectors of the same type to be used in the same application 
        this method must return a unique name for the connector. 
        A good practice is to return the name and a hash of the token.
        """
        h = hash_token(self.phone_number_id)
        return f"whatsapp:{h}"
    
    def set_gateway(self, gateway):
        from cel.gateway.message_gateway import MessageGateway
        assert isinstance(gateway, MessageGateway), \
            "gateway must be an instance of MessageGateway"
        self.gateway = gateway

    def get_router(self) -> APIRouter:
        return self.__build_routes(self.endpoint_prefix)

    def startup(self, context: MessageGatewayContext):
        log.warning(f"Be sure to setup in Meta Whatsapp Cloud API -> Webhoook this URL: {context.webhook_url}{self.endpoint_prefix}")

    def shutdown(self, context: MessageGatewayContext):
        pass

    def pause(self):
        self.paused = True

    def resume(self):
        self.paused = False
        

    async def __mark_as_read(self, message_id: str) -> dict:
        """Mark a message as read"""
        assert message_id is not None, "Message ID not provided"
        
        payload = {
          "messaging_product": "whatsapp",
          "status": "read",
          "message_id": message_id,
          "typing_indicator": { "type": "text" }
        }

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.post(f"{self.url}", 
                                    headers=build_headers(self.token), 
                                    json=payload) as response:
                if response.status == 200:
                    log.info(await response.json())
                    return await response.json()
                else:
                    log.error(await response.json())
                    return await response.json()              
        

    async def send_text_message(self, 
                                lead: WhatsappLead, 
                                text: str, 
                                metadata: dict = {}, 
                                is_partial: bool = True):
        """ Send a text message to the lead. The simplest way to send a message to the lead.
        
        Args:
            - lead[WhatsappLead]: The lead to send the message
            - text[str]: The text to send
            - metadata[dict]: Metadata to send with the message
            - is_partial[bool]: If the message is partial or not
        """
        
        assert isinstance(lead, WhatsappLead), "lead must be instance of WhatsappLead"
        
        data = {
            "messaging_product": "whatsapp",
            "recipient_type": "individual",
            "to": lead.phone,
            "type": "text",
            "text": {
                "preview_url": False, 
                "body": text
            },
        }
        
        headers = build_headers(self.token)
        log.info(f"Sending message to {lead.phone}")
        
        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.post(f"{self.url}", headers=headers, json=data) as r:
                if r.status == 200:
                    log.info(f"Message sent to {lead.phone}")
                else:
                    log.error(await r.json())

    async def send_select_message(self, 
                                  lead: WhatsappLead, 
                                  text: str,
                                  options: list[str],
                                  header: str = None,
                                  footer: str = None,
                                  button_text: str = None,
                                  list_title: str = None,
                                  metadata: dict = {}, 
                                  is_partial: bool = True):
        """ Send a select message to the lead. This will send a message with a list of options
        to the lead. The lead will be able to select one of the options.
        
        Args:
            - lead[WhatsappLead]: The lead to send the message
            - text[str]: The text to send with the options
            - options[list[str]]: A list of options to send to the lead
            - metadata[dict]: Metadata to send with the message
            - is_partial[bool]: If the message is partial or not
        """        
        assert isinstance(lead, WhatsappLead), "lead must be instance of WhatsappLead"
        assert options is not None, "Options not provided"
        assert len(options) > 0, "Options must be a list of strings"
        
        # trim options to 20 characters
        options = [opt[:20] for opt in options]
        
        if len(options) > 10:
            log.critical(f"The maximum number of options is 10. Got: {options}")
            # trim options to 10
            options = options[:10]
        
        items = [ListItem(title=opt) for opt in options]
        
        # "rows": [
        #     {
        #       "id": "<ROW_ID>",
        #       "title": "<ROW_TITLE_TEXT>",
        #       "description": "<ROW_DESCRIPTION_TEXT>"
        #     }
        #     /* Additional rows would go here*/
        #   ]
        
        response = await self._send_select(
            recipient_id=lead.phone,
            list={
                "type": "list",
                "header": {"type": "text", "text": header} if header else None,
                "body": {"text": text} if text else None,
                "footer": {"text": footer} if footer else None,
                "action": {
                    "sections": [
                        {
                            "title": list_title[:24] if list_title else None,
                            "rows": [json.loads(str(item)) for item in items]
                        }
                    ],
                    "button": button_text                    
                },

            }
        )
        log.debug(f"Select message response: {response}")
    
    
    async def send_buttons_message(self, 
                                  lead: WhatsappLead, 
                                  text: str,
                                  options: list[str],
                                  header: str = None,
                                  footer: str = None,
                                  metadata: dict = {}, 
                                  is_partial: bool = True):
        """ Send a message with buttons. This will send a message with a limited number of options
        to the lead. The lead will be able to select one of the options.
        For WhatsApp, the maximum number of buttons is 3. And the message is a interactive message with
        reply buttons.
        
        Args:
            - lead[WhatsappLead]: The lead to send the message
            - text[str]: The text to send with the options
            - options[list[str]]: A list of options to send to the lead
            - metadata[dict]: Metadata to send with the message
            - is_partial[bool]: If the message is partial or not
        """        
        assert isinstance(lead, WhatsappLead), "lead must be instance of WhatsappLead"
        assert options is not None, "Options not provided"
        assert len(options) > 0, "Options must be a list of strings"
        
        # trim options to 20 characters
        options = [opt[:20] for opt in options]
        
        if len(options) > 3:
            log.critical(f"The maximum number of options is 3. Got: {options}")
            # trim options to 3
            options = options[:3]
        
        buttons = [ReplyButton(title=opt) for opt in options]
        
        response = await self._send_reply_button(
            recipient_id=lead.phone,
            button={
                "type": "button",
                "header": {"type": "text", "text": header} if header else None,
                "body": {"text": text} if text else None,
                "footer": {"text": footer} if footer else None,
                "action": {
                    "buttons": [json.loads(str(button)) for button in buttons]
                }
            }
        )
        log.debug(f"Select message response: {response}")    
    
    async def send_link_message(self, lead: WhatsappLead, text: str, links: list, metadata: dict = {}, is_partial: bool = True):
        """ Send a link message to the lead. When you need to send a link to some platforms
        like WhatsApp, the platform will automatically generate a preview of the link and embed a link
        to the text in a normal message.
        This method will help you to send a link message most effectively, hidden the preview and 
        embeding the link into a button. This will bring a better user experience.
        
        Args:
        
            - lead[WhatsappLead]: The lead to send the message
            - text[str]: The text to send with the link
            - links[list]: A list of dictionaries containing the link data. 
            Each dictionary must have a `text` and `url` key.
            - metadata[dict]: Metadata to send with the message
            - is_partial[bool]: If the message is partial or not
    
        Example:
            
            ```python
            links = [
                {"text": "Go to Google", "url": "https://www.google.com"},
                {"text": "Go to Facebook", "url": "https://www.facebook.com"}
            ]
            await conn.send_link_message(ctx.lead, text="Please follow this link", links=links)
            ```
    
        """
        assert isinstance(lead, WhatsappLead), "lead must be instance of WhatsappLead"
        assert links is not None, "Links not provided"
        assert len(links) > 0, "Links must be a list of dictionaries"
        assert text is not None, "Text not provided"
        
        log.debug(f"Sending Link Message to phone: {lead.phone}, text: {text}, url: {links}")
        response = await self._send_link(
            link_label=links[0].get("text"),
            link=links[0].get("url"),
            body=text,
            recipient_id=lead.phone,
        )
         
    async def send_typing_action(self, lead):
        """Send typing action to the lead. This will show the typing action in the chat.
        
        Args:
            - lead[TelegramLead]: The lead to send the typing action
        """    
        assert isinstance(lead, WhatsappLead), "lead must be instance of WhatsappLead"
        # do nothing, there is no typing action in whatsapp cloud api
        log.warning(f"Whatsapp typing action not currently supported by Cloud API")
        pass
    
    
    async def send_message(self, message: OutgoingMessage):
        assert isinstance(message, OutgoingMessage),\
            "message must be an instance of OutgoingMessage"
        assert isinstance(message.lead, WhatsappLead),\
            "lead must be an instance of TelegramLead"
        lead = message.lead
        
        if message.type == OutgoingMessageType.TEXT:
            assert isinstance(message, OutgoingTextMessage),\
            "message must be an instance of OutgoingMessage"
            await self.send_text_message(lead, 
                                         message.content, 
                                         metadata=message.metadata, 
                                         is_partial=message.is_partial)
            
        if message.type == OutgoingMessageType.SELECT:
            assert isinstance(message, OutgoingSelectMessage),\
            "message must be an instance of OutgoingSelectMessage"
            
            await self.send_select_message(lead, 
                                           message.content, 
                                           options=message.options, 
                                           metadata=message.metadata, 
                                           button_text=message.button,
                                           list_title=message.list_title,
                                           is_partial=message.is_partial)
            
        if message.type == OutgoingMessageType.BUTTONS:
            assert isinstance(message, OutgoingButtonsMessage),\
            "message must be an instance of OutgoingSelectMessage"
            
            await self.send_buttons_message(lead, 
                                           message.content, 
                                           options=message.options, 
                                           metadata=message.metadata, 
                                           is_partial=message.is_partial)            
        
        if message.type == OutgoingMessageType.LINK:
            assert isinstance(message, OutgoingLinkMessage),\
            "message must be an instance of OutgoingLinkMessage"
            
            await self.send_link_message(lead, 
                                         message.content, 
                                         links=message.links, 
                                         metadata=message.metadata, 
                                         is_partial=message.is_partial)


    async def _send_link(self, 
                         link_label: str, 
                         body: str | None,
                         link: str,
                         recipient_id: str,
                         footer: str | None = None,
                         image_url: str = None) -> None:
        await self._send_cta_url(
            recipient_id=recipient_id,
            cta_url={
                "type": "cta_url",
                "header": {
                    "type": "image",
                    "image": {"link": image_url}
                } if image_url else None,
                "body": {
                    "text": body
                },
                "footer": {
                    "text": footer
                },
                "action": {
                    "name": "cta_url",
                    "parameters": {
                        "display_text": link_label,
                        "url": link
                    }
                }                
            }
        )

    def _create_button(self, button: Dict[Any, Any]) -> Dict[Any, Any]:
        """
        Method to create a button object to be used in the send_message method.

        Args:
                button[dict]: A dictionary containing the button data
        """
        data = {"type": "list", "action": button.get("action")}
        if button.get("header"):
            data["header"] = {"type": "text", "text": button.get("header")}
        if button.get("body"):
            data["body"] = {"text": button.get("body")}
        if button.get("footer"):
            data["footer"] = {"text": button.get("footer")}
        return data

    async def _send_button(self, button: Dict[Any, Any], recipient_id: str) -> Dict[Any, Any]:
        """
        Sends an interactive buttons message to a WhatsApp user

        Args:
            button[dict]: A dictionary containing the button data(rows-title may not exceed 20 characters)
            recipient_id[str]: Phone number of the user with country code wihout +

        check https://github.com/Neurotech-HQ/whatsapp#sending-interactive-reply-buttons for an example.
        """
        assert isinstance(recipient_id, str) and recipient_id != "", "Recipient ID must be a string"
        assert isinstance(button, dict), "Button must be a dictionary"
        
        data = {
            "messaging_product": "whatsapp",
            "to": recipient_id,
            "type": "interactive",
            "interactive": self._create_button(button),
        }
        
        log.debug(f"Sending buttons to {recipient_id}")
        headers = build_headers(self.token)
        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.post(f"{self.url}", headers=headers, json=data) as r:
                if r.status == 200:
                    log.debug(f"Message sent to {recipient_id}")
                else:
                    log.error(await r.json())     
                    
    async def _send_select(
        self, 
        list: Dict[Any, Any], 
        recipient_id: str
    ) -> Dict[Any, Any]:
        """
        Sends an interactive list message to a WhatsApp user

        Args:
            button[dict]: A dictionary containing the button data
            recipient_id[str]: Phone number of the user with country code wihout +

        Note:
            The maximum number of buttons is 3, more than 3 buttons will rise an error.
        """

        
        data = {
            "messaging_product": "whatsapp",
            "recipient_type": "individual",
            "to": recipient_id,
            "type": "interactive",
            "interactive": list,
        }
        headers = build_headers(self.token)
        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.post(f"{self.url}", headers=headers, json=data) as r:
                if r.status == 200:
                    log.debug(f"Message sent to {recipient_id}")
                else:
                    log.error(await r.json())                       
                       

    async def _send_reply_button(
        self, 
        button: Dict[Any, Any], 
        recipient_id: str
    ) -> Dict[Any, Any]:
        """
        Sends an interactive reply buttons[menu] message to a WhatsApp user

        Args:
            button[dict]: A dictionary containing the button data
            recipient_id[str]: Phone number of the user with country code wihout +

        Note:
            The maximum number of buttons is 3, more than 3 buttons will rise an error.
        """
        if len(button["action"]["buttons"]) > 3:
            raise ValueError("The maximum number of buttons is 3.")
        
        data = {
            "messaging_product": "whatsapp",
            "recipient_type": "individual",
            "to": recipient_id,
            "type": "interactive",
            "interactive": button,
        }
        headers = build_headers(self.token)
        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.post(f"{self.url}", headers=headers, json=data) as r:
                if r.status == 200:
                    log.debug(f"Message sent to {recipient_id}")
                else:
                    log.error(await r.json())     
                    
    async def _send_cta_url(
        self, cta_url: Dict[Any, Any], 
        recipient_id: str
    ) -> Dict[Any, Any]:
        """
        Sends a call to action url message to a WhatsApp user

        Args:
            cta_url[dict]: A dictionary containing the cta url data
            recipient_id[str]: Phone number of the user with country code wihout +

        check
        
        """
        data = {
            "messaging_product": "whatsapp",
            "recipient_type": "individual",
            "to": recipient_id,
            "type": "interactive",
            "interactive": cta_url,
        }
        headers = build_headers(self.token)
        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.post(f"{self.url}", headers=headers, json=data) as r:
                if r.status == 200:
                    log.debug(f"Message sent to {recipient_id}")
                else:
                    log.error(await r.json())     



    async def send_document_message(self, lead, content: str | bytes | BinaryIO, 
                                    filename:str, caption:str = None, metadata: dict = {}, is_partial: bool = True):
        log.debug(f"Sending document {filename} (caption:{caption}) message to {lead.phone}")
        
        assert isinstance(lead, WhatsappLead), "lead must be instance of WhatsappLead"
        assert isinstance(filename, str), "filename must be a string"
        
        from pywa import WhatsApp
        
        # get mimet type from filename
        import mimetypes
        mime_type, _ = mimetypes.guess_type(filename)
        log.debug(f"Mime type: {mime_type}")

        wa = WhatsApp(
            phone_id=self.phone_number_id,
            token=self.token
        )

        # Send the document
        wa.send_document(
            to=lead.phone,
            document=content,
            filename=filename,
            caption=caption,
            mime_type=mime_type,
            footer=caption
        )



    # all the files starting with _ are imported here, and should not be imported directly.
    changed_field = staticmethod(changed_field)
    is_message = staticmethod(is_message)
    is_reaction = staticmethod(is_reaction)



from abc import ABC
import re
import time
from typing import Optional
from fastapi import HTTPException
from pydantic import BaseModel
from redis import asyncio as aioredis
import json
from dataclasses import asdict, dataclass
from cel.assistants.base_assistant import BaseAssistant
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.message import Message
from loguru import logger as log

Redis = aioredis.Redis

# WARNING: DO NOT USE THIS KEY IN PRODUCTION
DEFUALT_MASTER_KEY = "123456"


@dataclass
class AuthEntry(ABC):
    session_id: str = None
    invite_code: str = None
    client_cmd_enabled: bool = False
    last_request: int = 0
    metadata: dict = None
    
@dataclass
class InvitationEntry(ABC):
    invite_code: str = None
    created_at: int = 0
    # 0 means never expires
    expires_at: int = 0
    used: bool = False
    name: str = None
    metadata: dict = None

class InvitationGuardMiddlewareEvents:
    new_conversation = "new_conversation"
    invitation_accepted = "invitation_accepted"
    rejected_code = "rejected_code"
    admin_login = "admin_login"
    admin_logout = "admin_logout"



class InvitationGuardMiddleware(ABC):
    """Middleware to secure client commands, session handling and user lyfe cycle related events
    Invite code is a 6 alphanumeric characters, including uppercase and lowercase.
    This middleware will search for the invite code in the message and handle it accordingly.
    It specs code formatted as #XXXXXX where X is an alphanumeric character.
    
    
    Invitations:
        - Create, revoke and get invitations
        - Claim an invitation
        - Handle invitation codes in messages
        - Backdoor code to bypass invitation system
        
    
    Triggered events:
        - new_conversation: Event called when a new conversation is started.
        - rejected_code: Event called when a user enters an invalid code.
        - admin_login: Event called when a user logs in as admin.
        - admin_logout: Event called when a user logs out as admin.
        
    Args:
        telegram_bot_name: The name of the Telegram bot for URL generation
        whatsapp_phone_number: The phone number of the WhatsApp bot for URL generation
        redis: The Redis database URL
        key_prefix: The prefix for the keys in the Redis database
        master_key: The master key to login as admin
        reject_message: The message to send when a user is rejected
        backdoor_invite_code: The backdoor invite code
        allow_only_invited: Allow only invited users to access the service
        
    """
    
    events: InvitationGuardMiddlewareEvents = InvitationGuardMiddlewareEvents()
    
    
    def __init__(self, 
                 telegram_bot_name: str = None,
                 whatsapp_phone_number: str = None,
                 redis: str | Redis = None, 
                 key_prefix: str = "igm", 
                 master_key: str = None,
                 # default reject message with wrong way emoji
                 reject_message: str = "🚫 You are not allowed to access this service.",
                 backdoor_invite_code: str = None,
                 allow_only_invited: bool = True):
        
        self.client = redis if isinstance(redis, Redis) else aioredis.from_url(redis or 'redis://localhost:6379/0')
        self.key_prefix = key_prefix
        log.critical("No master key provided. Using default key") if not master_key else None
        self.master_key = master_key or DEFUALT_MASTER_KEY
        self.allow_only_invited = allow_only_invited
        self.reject_message = reject_message
        self.backdoor_invite_code = backdoor_invite_code
        self.telegram_bot_name = telegram_bot_name
        self.whatsapp_phone_number = whatsapp_phone_number
        
        
    
    # Setup the middleware with the FastAPI application    
    def setup(self, app):
        """ Setup the middleware with the FastAPI application. Here you can define
        the routes for the invitation system.
        
        Use the prefix /middleware/guard to register the routes.
        """
        log.debug("InvitationGuardMiddleware initialized")
        prefix = "/middleware/guard"
        
        class InvitationRequest(BaseModel):
            # code: str
            expires_at: Optional[int] = 0
            name: Optional[str] = None

        class InvitationResponse(BaseModel):
            invite_code: str
            message: str
        
        # Add echo route
        @app.get(prefix + "/echo")
        async def echo():
            return {"status": "ok"}
        
        @app.get(prefix + "/invitations/{code}", 
                 response_model=InvitationEntry, 
                 summary="Get invitation", 
                 description="This endpoint allows you to get an invitation by code.")
        async def get_invitation(code: str):
            """
            Get invitation by code
            """
            entry = await self.get_invitation(code)
            
            if entry is None:
                raise HTTPException(status_code=404, detail="Invitation not found")
            
            return entry 
        
        @app.post(prefix + "/invitations", 
                  response_model=InvitationResponse, 
                  summary="Create a new invitation", 
                  description="This endpoint allows you to create a new invitation.")
        async def create_invitation(request: InvitationRequest):
            """
            Create a new invitation

            - **code**: The invitation code
            - **expires_at**: The expiration time of the invitation (optional)
            - **name**: The name associated with the invitation (optional)
            """
            code = self.__gen_invite_code()
            await self.create_invitation(code, name=request.name, expires_at=request.expires_at)
            return {
                "message": "Invitation created successfully",
                "invite_code": code
            }
        
        @app.delete(prefix + "/invitations/{code}", 
                    response_model=InvitationResponse, 
                    summary="Revoke an invitation", 
                    description="This endpoint allows you to revoke an invitation.")
        async def revoke_invitation(code: str):
            """
            Revoke an invitation by code
            """
            await self.revoke_invitation(code)
            return {"message": "Invitation revoked successfully"}

    # MAIN METHOD - Handle the message
    async def __call__(self, message: Message, connector: BaseConnector, assistant: BaseAssistant):
        try:
            assert isinstance(message, Message), "Message must be a Message object"
            assert isinstance(connector, BaseConnector), "Connector must be a BaseConnector object"
          
            if not await self.__handle_invitation_code(message, connector, assistant):
                return False        
            
            entry = await self.get_auth_entry(message.lead.get_session_id())
            
            # add last request time in message
            if entry:
                assert isinstance(entry, AuthEntry), "Entry must be an AuthEntry object"
                
                if self.allow_only_invited and entry.invite_code is None:
                    log.critical(f"User {message.lead.get_session_id()} is not invited")
                    await connector.send_text_message(message.lead, self.reject_message)
                    return False
                
                time_since_last_request = time.time() - (entry.last_request or 0)
                message.metadata = message.metadata or {}
                message.metadata['time_since_last_request'] = time_since_last_request
                message.metadata['invitation'] = entry.metadata
                message.lead.metadata = message.lead.metadata or {}
                message.lead.metadata['time_since_last_request'] = time_since_last_request
                message.lead.metadata['invitation'] = entry.metadata
                
                await self.set_entry(message.lead.get_session_id(), 
                                    client_cmd_enabled=entry.client_cmd_enabled, 
                                    metadata=entry.metadata,
                                    invite_code=entry.invite_code)
            else:
                await self.set_entry(message.lead.get_session_id(), 
                                    client_cmd_enabled=False, 
                                    metadata={})
                await connector.send_text_message(message.lead, self.reject_message)
                await assistant.call_event(self.events.new_conversation, message.lead, message, connector)
                return False
                
            if not await self.__handle_login_command(message, connector, entry, assistant):
                return False
            
            if not await self.__secure_client_commands(message, connector, entry):
                return False
            
            return True
        except Exception as e:
            log.error(f"Error in InvitationGuardMiddleware: {e}")
            return False
    
    
    # Handle invitation claim
    def __search_invitation_code(self, text: str):
        if not text:
            return None        
        
        # regex to match code in string
        regex = r"#\w{6}"
        
        # search for the pattern in the message content
        match = re.search(regex, text)
        
        # return the matched code or None if no match is found
        if match:
            return match.group(0)
        
        return None
    
    async def __handle_invitation_code(self, message: Message, connector: BaseConnector, assistant: BaseAssistant):
        code = self.__search_invitation_code(message.text)
        if code:
            inv = await self.get_invitation(code)
            
            if code == self.backdoor_invite_code and self.backdoor_invite_code:
                await self.set_entry(message.lead.get_session_id(), invite_code=code)
                await connector.send_text_message(message.lead, "Backdoor code accepted")
                message.text = message.text.replace(code, "")
                await assistant.call_event(self.events.invitation_accepted, message.lead, message, connector, data=inv)
                return True
            
            # Reject cases
            # --------------------------------
            if inv is None:
                log.critical(f"Invalid code {code} provided")
                await connector.send_text_message(message.lead, "Invalid code provided")
                await assistant.call_event(self.events.rejected_code, message.lead, message, connector)
                return False
            
            if inv.expires_at > 0 and time.time() > inv.expires_at:
                log.critical(f"Expired code {code} provided")
                await connector.send_text_message(message.lead, "Expired code provided")
                await assistant.call_event(self.events.rejected_code, message.lead, message, connector)
                return False
            
            if inv.used:
                log.critical(f"Code {code} already used")
                await connector.send_text_message(message.lead, "Code already used")
                await assistant.call_event(self.events.rejected_code, message.lead, message, connector)
                return False

            # Accept the code
            # --------------------------------
            await self.set_entry(message.lead.get_session_id(), invite_code=code, metadata=inv.metadata)
            await self.claim_invitation(code)
            await connector.send_text_message(message.lead, "Code accepted")      
            message.text = message.text.replace(code, "")      
            await assistant.call_event(self.events.invitation_accepted, message.lead, message, connector, data=inv)
            
            return True
        return True
                
    # Handle invitations 
    def __gen_invite_code(self):
        # genereate a random code 6 alphanumeric characters, including uppercase and lowercase
        import random
        import string
        code = '#' + ''.join(random.choices(string.ascii_uppercase + string.ascii_lowercase + string.digits, k=6))
        return code
    
    def __gen_barcode(self, url: str):
        import qrcode
        import io
        log.debug(f"Generating QR code for {url}")
        qr = qrcode.make(url)
        stream = io.BytesIO()
        qr.save(stream)
        qr_bytes = stream.getvalue()
        return qr_bytes
    
    def __gen_telegram_invitation_url(self, code: str):
        assert self.telegram_bot_name, "Telegram bot name must be provided"
        import base64
        #  the code must be up to 64 base64url characters
        input_bytes = code.encode('utf-8')
        base64url_bytes = base64.urlsafe_b64encode(input_bytes)
        code_encoded = base64url_bytes.decode('utf-8')        
        
        assert len(code_encoded) <= 64, "Code must be up to 64 base64url characters"
    
        url = f"https://t.me/{self.telegram_bot_name}?start={code_encoded}"
        return url        
        
    def __gen_whatsapp_invitation_url(self, code: str):
        assert self.whatsapp_phone_number, "WhatsApp phone number must be provided"
        url = f"https://wa.me/{self.whatsapp_phone_number}?text={code}"
        return url
    
    async def create_invitation(self, name: str = None, metadata: dict = None, expires_at: int = 0):
        code = self.__gen_invite_code()
        entry = InvitationEntry(invite_code=code, 
                                created_at=int(time.time()), 
                                expires_at=expires_at, 
                                metadata=metadata,
                                name=name)
        await self.client.hset(self.key_prefix, code, json.dumps(asdict(entry)))
        return entry
    
    async def get_invitation(self, code: str):
        entry = await self.client.hget(self.key_prefix, code)
        if entry:
            entry = json.loads(entry)
            return InvitationEntry(invite_code=entry.get('invite_code'), 
                                   created_at=entry.get('created_at'), 
                                   expires_at=entry.get('expires_at'),
                                   used=entry.get('used'), 
                                   metadata=entry.get('metadata'),
                                   name=entry.get('name'))
        return None
    
    async def claim_invitation(self, code: str):
        inv = await self.get_invitation(code)
        if inv:
            inv.used = True
            await self.client.hset(self.key_prefix, code, json.dumps(asdict(inv)))
            return inv
        return None
    
    async def revoke_invitation(self, code: str):
        await self.client.hdel(self.key_prefix, code)

    async def clear_invitations(self):
        log.warning(f"Clearing invitations from Redis, all keys with prefix {self.key_prefix} will be deleted")
        await self.client.delete(self.key_prefix)
        log.debug("Invitations cleared successfully")
    
    # Handle login/logout commands and secure client commands
    async def __handle_login_command(self, 
                                     message: Message, 
                                     connector: BaseConnector, 
                                     entry: AuthEntry,
                                     assistant: BaseAssistant):
        text = message.text or ''
        if text.startswith("/logout"):
            await self.clear_auth(message.lead.get_session_id())
            await connector.send_text_message(message.lead, "You have been logged out!")
            await assistant.call_event(self.events.admin_logout,  message.lead, message, connector)
            return False
        
        if text.startswith("/login"):
            parts = text.split(" ")
            if len(parts) == 2:
                if parts[1] == self.master_key:
                    await self.enable_client_commands(message.lead.get_session_id())
                    await assistant.call_event(self.events.admin_login,  message.lead, message, connector)
                    await connector.send_text_message(message.lead, "You have been logged in.")
                    await connector.send_text_message(message.lead, "Welcome AI bender!")
                    return False
                else:
                    log.critical(f"Invalid master key provided for login")
                    return False
            else:
                log.critical(f"Invalid login command format")
                return False
        return True
        
    async def __secure_client_commands(self, message: Message, connector: BaseConnector, entry: AuthEntry):
        text = message.text or ''
        if text.startswith("/") and not text.startswith("/login") and not text.startswith("/logout"):
            if not entry or not entry.client_cmd_enabled:
                log.critical(f"Client command {text} from {message.lead.get_session_id()} is not allowed")
                return False

        if text.startswith("/help"):
            await connector.send_text_message(message.lead, "Available commands:")
            await connector.send_text_message(message.lead, "/login <master_key> - Login as admin")
            await connector.send_text_message(message.lead, "/logout - Logout as admin")
            await connector.send_text_message(message.lead, "/help - Show help")
            await connector.send_text_message(message.lead, "/reset_auth - Reset session data")
            await connector.send_text_message(message.lead, "/reset_to_fabric - Reset session data and revoke all invitations")
            await connector.send_text_message(message.lead, "/invite <name> - Create an invitation")
            await connector.send_text_message(message.lead, "/authinfo - Show session data")
        
        if text.startswith("/reset_to_fabric"):
            # clear entry for this session
            await connector.send_text_message(message.lead, "Resetting session data")
            await self.clear_auth(message.lead.get_session_id())
            await connector.send_text_message(message.lead, "Revoking all invitations!")
            await self.clear_invitations()
            await connector.send_text_message(message.lead, "Now you are uninvited, logged out and all invitations are revoked")
            
        if text.startswith("/reset_auth"):
            # clear entry for this session
            await connector.send_text_message(message.lead, "Resetting session data")
            await self.clear_auth(message.lead.get_session_id())
            
        if text.startswith("/authinfo"):
            await connector.send_text_message(message.lead, f"Auth info: {entry}")
            await connector.send_text_message(message.lead, f"Session ID: {message.lead.get_session_id()}")
        
        if text.startswith("/invite"):
            parts = text.split(" ")
            if len(parts) == 2:
                invitation = await self.create_invitation(name=parts[1])
                source = message.lead.connector_name
                qr, url = await self.get_invitation_assets(message.lead, invitation)
                    
                if qr and url:
                    await connector.send_image_message(message.lead, qr, filename="qrcode.png")
                    await connector.send_text_message(message.lead, f"Invitation link: {url}")
                    await connector.send_text_message(message.lead, f"Code: {invitation.invite_code}")
                else: 
                    log.error(f"Invalid source {source} for invitation")
            else:
                await connector.send_text_message(message.lead, "Invalid invite command format")
            
            
        return True
    
    async def get_invitation_assets(self, lead: ConversationLead, invitation: InvitationEntry):
        qr = None
        url = None
        source = lead.connector_name
        
        if source.startswith("telegram:"):
            url = self.__gen_telegram_invitation_url(invitation.invite_code)
            qr = self.__gen_barcode(url)
        
        if source.startswith("whatsapp:"):
            url = self.__gen_whatsapp_invitation_url(invitation.invite_code)
            qr = self.__gen_barcode(url)
        
        return qr, url
    
    async def send_invitation_assets(self, lead: ConversationLead, invitation: InvitationEntry):
        qr, url = await self.get_invitation_assets(lead, invitation)
        connector = lead.connector
        await connector.send_image_message(lead, qr, filename="qrcode.png")
        await connector.send_text_message(lead, f"Invitation link: {url}")
        await connector.send_text_message(lead, f"Code: {invitation.invite_code}")


    # Auth Methods
    async def set_entry(self, 
                        session_id: str, 
                        client_cmd_enabled: bool = False, 
                        metadata: dict = None,
                        invite_code: str = None):
        
        entry = AuthEntry(session_id=session_id, 
                          client_cmd_enabled=client_cmd_enabled, 
                          metadata=metadata,
                          invite_code=invite_code,
                          last_request=time.time()
                        )
        await self.client.hset(self.key_prefix, session_id, json.dumps(asdict(entry)))
        return entry
        
    async def clear_auth(self, id: str):
        await self.client.hdel(self.key_prefix, id)
        
    async def get_auth_entry(self, id: str):
        entry = await self.client.hget(self.key_prefix, id)
        if entry:
            entry = json.loads(entry)
            return AuthEntry(session_id=entry.get('session_id'),
                            invite_code=entry.get('invite_code'),
                            client_cmd_enabled=entry.get('client_cmd_enabled'), 
                            last_request=entry.get('last_request'),
                            metadata=entry.get('metadata'))

        return None
    
    async def enable_client_commands(self, id: str):
        entry = await self.get_auth_entry(id)
        if entry:
            entry.client_cmd_enabled = True
            await self.client.hset(self.key_prefix, id, json.dumps(asdict(entry)))
            return entry
        return None


import json
from abc import ABC

class ConversationPeer(ABC):
    
    def __init__(self, name: str, metadata: dict = None, id: str = None, phone: str = None, avatarUrl: str = None, email: str = None):
        self.name = name
        self.metadata = metadata
        self.id = id
        self.phone = phone
        self.avatarUrl = avatarUrl
        self.email = email
        
    def to_dict(self):
        return {
            'name': self.name,
            'metadata': self.metadata,
            'id': self.id,
            'phone': self.phone,
            'avatarUrl': self.avatarUrl,
            'email': self.email
        }
        
    def to_json(self):
        return json.dumps(self.to_dict())
    
    def __str__(self):
        return f"ConversationPeer: {self.name}"
    
    @classmethod
    def from_dict(cls, peer_dict):
        return ConversationPeer(
            name=peer_dict.get("name"),
            metadata=peer_dict.get("metadata"),
            id=peer_dict.get("id"),
            phone=peer_dict.get("phone"),
            avatarUrl=peer_dict.get("avatarUrl"),
            email=peer_dict.get("email")
        )

class ConversationPeerEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, ConversationPeer):
            return obj.to_dict()
        return super().default(obj)


# if __name__ == "__main__":
#     peer = ConversationPeer(name="John Doe")
    
#     obj = {
#         "id": 1,
#         "peer": peer
#     }

#     # json.dumps(obj)

#     json_str = json.dumps(obj, cls=ConversationPeerEncoder)
#     print(json_str)
    


from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.conversation_peer import ConversationPeer


class LiveKitLead(ConversationLead):
    """ LiveKitLead class."""

    def __init__(self, call_object: dict, **kwargs):
        super().__init__(connector_name="livekit", **kwargs)
        self.call_object: dict = call_object

    def get_session_id(self):
        return f"{self.connector_name}:{self.call_object['session_id']}"
    
    @classmethod
    def from_message(cls, message: dict, **kwargs):
        conversation_peer = ConversationPeer(
            name='Unknown',
            id=message['session_id'],
            phone=None,
            avatarUrl=None,
            email=None,
        )
        return LiveKitLead(call_object=message, conversation_from=conversation_peer, **kwargs)



from cel.gateway.model.conversation_lead import ConversationLead
from cel.stores.history.base_history_provider import BaseHistoryProvider
from langchain.load.load import load



async def build_router_query (history_store: BaseHistoryProvider, lead: ConversationLead, text: str, length: int = 5):
    history = await history_store.get_history(lead.get_session_id())
    # Create a list of the last N messages without tools and tool_calls
    messages = []
    for h in history:
        aux = load(h)
        role = aux.type 
        text = aux.content
        
        # Skip tools and tool_calls
        if role == "tool" or role == "tool_call":
            continue
        
        messages.append({
            "role": role,
            "text": text
        })

    # Get the last N messages
    last_messages = messages[-length:]
    return last_messages






from dataclasses import dataclass

ChatwootMessageTypes = ["incoming", "outgoing"]

@dataclass
class ChatwootConversationRef:
    id: int
    identifier: str
    updated_at: int
    assigned: bool = False
    
    
@dataclass
class InboxRef:
    id: int
    name: str
    
@dataclass
class ContactLead:
    identifier: str
    celai_lead: str
    name: str = None
    email: str = None
    phone_number: str = None
    
    



from abc import ABC
import inspect
import json

import shortuuid
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.conversation_peer import ConversationPeer, ConversationPeerEncoder


class ConversationLead(ABC):

    def __init__(self, 
                 metadata : dict = None, 
                 connector_name: str = None,
                 conversation_from: ConversationPeer = None,
                 connector: BaseConnector = None):
        self.metadata = metadata
        self.conversation_from = conversation_from
        self.tmp_id = shortuuid.uuid()
        self.connector = connector
        self.connector_name = connector.name() if connector else connector_name or None

    def get_session_id(self):
        """ This method should be overriden in the child class """
        return self.tmp_id

    def to_dict(self):
        return {
            'connector_name': self.connector_name,
            'metadata': self.metadata,
            'conversation_from': self.conversation_from.to_dict() if self.conversation_from else None
        }

    def to_json(self):
        return json.dumps(self.to_dict())
    
    @classmethod
    def from_dict(cls, lead_dict):
        # connector_name=lead_dict.get("connector_name")
        return ConversationLead(
            metadata=lead_dict.get("metadata", {}),
            conversation_from=ConversationPeer.from_dict(lead_dict.get("conversation_from", {}) or {})
        )
        
    def __str__(self):
        return f"ConversationLead: {self.connector_name}"
    

    @staticmethod
    def serialize(lead: 'ConversationLead') -> str:
        """ Serialize a ConversationLead object to JSON. If you want to take this
        this object outside the process, you can use this method to serialize it
        It allows you to serialize the object and then deserialize it in another process"""
        
        # Obtener el nombre de la clase
        class_name = f"{lead.__class__.__module__}.{lead.__class__.__name__}"
        # Obtener los atributos del objeto
        attributes = {k: v for k, v in lead.__dict__.items() if k != 'metadata'}
        #  remove connector from attributes
        attributes.pop('connector', None)
        # Crear un diccionario con el nombre de la clase y los atributos
        data = {
            'class_name': class_name,
            'attributes': attributes
        }
        # Serializar a JSON
        return json.dumps(data, cls=ConversationPeerEncoder)
    
    
    @staticmethod
    def deserialize(serialized_data: str) -> 'ConversationLead':
        """ Deserialize a JSON string to a ConversationLead object. 
        and re-attach the source connector to the object attribute connector."""
        
        # Deserializar el JSON
        data = json.loads(serialized_data)
        class_name = data['class_name']
        attributes = data['attributes']
        
        import importlib

        def dynamic_class_instantiation(class_path: str):
            module_path, class_name = class_path.rsplit('.', 1)
            module = importlib.import_module(module_path)
            cls = getattr(module, class_name)
            return cls
        
        def get_connector_instance(connector_name: str):
            if connector_name:
                return BaseConnector.get_connector_by_name(connector_name)
            return None   
        
        # Obtener la clase a partir del nombre
        cls = dynamic_class_instantiation(class_name)
        
        # Inspeccionar el constructor de la clase
        constructor_params = inspect.signature(cls).parameters
        
        # Filtrar los atributos que coinciden con los parámetros del constructor
        init_args = {k: attributes[k] for k in constructor_params if k in attributes}
        
        # Crear una instancia de la clase con los argumentos
        instance = cls(**init_args)
        
        # set attributes
        for k, v in attributes.items():
            if k not in constructor_params:
                setattr(instance, k, v)
        
        # set connector
        instance.connector = get_connector_instance(instance.connector_name)
        
        return instance


from abc import ABC
from .component import Component
import json
from .utils import del_none

class ReplyButton(Component):
    def __init__(self, title: str, id: str = None) -> None:
        self.title = title
        self.id = id or title
        
    def __str__(self):
        button = {
            "type": "reply",
            "reply": {"id": self.id, "title": self.title }
        }
        del_none(button)
        return json.dumps(button)
        
    
    def __repr__(self):
        return f"ReplyButton(id={self.id}, title={self.title})"
        
    
    def __eq__(self, other):
        return self.id == other.id and self.title == other.title


if __name__ == "__main__":
    button = ReplyButton("id", "title")
    print(button)
    # Expected output: '{"type": "reply", "reply": {"id": "id", "title": "title"}}'


#  define roles type
from abc import ABC
from dataclasses import dataclass

from cel.assistants.common import FunctionDefinition



class Role:
    USER = "user"
    ASSISTANT = "assistant"
    SYSTEM = "system"
    FUNCTION = "function"
    
@dataclass
class ContextMessage(ABC):
    role: Role
    content: str
    name: str = None
    function_call: dict = None
    
    
@dataclass
class PromptCompiled:
    prompt: str
    settings: dict
    functions: list[FunctionDefinition]
    context: dict
    
@dataclass
class PromptContext:
    history: list[ContextMessage]
    state: dict
    message: str | None
    date: str





from diskcache import Cache
from .base_cache import BaseCache

class DiskCache(BaseCache):
    def __init__(self, cache_dir: str):
        self.cache = Cache(cache_dir)

    def memoize(self, typed: bool, expire: int, tag: str):
        return self.cache.memoize(typed=typed, expire=expire, tag=tag)


import asyncio
from cel.middlewares.chatwoot.chatwoot_client import ChatwootClient
from cel.middlewares.chatwoot.model import ChatwootConversationRef, ContactLead, InboxRef
from cel.middlewares.chatwoot.phone_util import format_to_e164
from loguru import logger as log
import time
from typing import Any, Dict, Optional




class ConversationManager:

    def __init__(self,
                 base_url: str,
                 access_key: str,
                 webhook_url: str,
                 account_id: str,
                 inbox_name: str):
        self.base_url = base_url
        self.access_key = access_key
        self.inbox_name = inbox_name
        self.account_id = account_id
        self.webhook_url = webhook_url
        self.client = ChatwootClient(base_url, account_id, access_key)
        self.inbox = None
        self.conversations = {}

    def __set_inbox(self, inbox):
        self.inbox = InboxRef(
            id=inbox.get("id"),
            name=inbox.get("name")
        )
        log.debug(f"Chawoot Inbox: {self.inbox}")


    async def init(self, auto_create_inbox: bool = True):
        """ Initialize the ConversationManager 
            auto_create_inbox: If True, create the inbox if not found.
        """
        log.debug(f"Initializing Chatwoot ConversationManager: {self.base_url}, account_id: {self.account_id}, inbox_name: {self.inbox_name}")
        # 1. Get inbox by name
        # 2. If not found, create inbox
        inbox = await self.client.get_inbox_by_name(self.account_id, self.inbox_name)
        
        # check if webhook url needs to be updated
        if inbox and inbox.get("webhook_url") != self.webhook_url:
            log.warning(f"Updating webhook url for inbox: {self.inbox_name} in Chatwoot: {self.base_url}")
            inbox = await self.client.update_api_inbox(self.account_id, inbox.get("id"), self.webhook_url)
            log.warning(f"Webhook url updated for inbox: {self.inbox_name} in Chatwoot: {self.base_url}")

        if not inbox and not auto_create_inbox:
            raise Exception(f"Inbox {self.inbox_name} not found in Chatwoot: {self.base_url}")

        if not inbox:
            log.warning(f"Inbox {self.inbox_name} not found in Chatwoot: {self.base_url}")
            inbox = await self.client.create_api_inbox(self.account_id, self.inbox_name, self.webhook_url)
            if not inbox:
                raise Exception(f"Failed to create inbox {self.inbox_name} in Chatwoot: {self.base_url}")

        self.__set_inbox(inbox)



    async def get_conversation(self, identifier: str) -> Optional[ChatwootConversationRef]:
        """ Get a conversation by Contact identifier 
            Conversations are stored in memory for faster access.
            Index by identifier
        """
        return self.conversations.get(identifier)
    
    def get_conversation_by_id(self, conversation_id: int) -> Optional[ChatwootConversationRef]:
        """ Get a conversation by Chatwoot conversation ID
            Conversations are stored in memory for faster access.
            Index by identifier
        """
        for conv in self.conversations.values():
            if conv.id == conversation_id:
                return conv
        return None

    async def create_new_conversation(self, 
                                      contact_id: int, 
                                      source_id: int, 
                                      first_message: str = None,
                                      custom_attributes: Dict[str, str] = None) -> Optional[ChatwootConversationRef]:
        inbox_id = self.inbox.id
        conversation = await self.client.create_new_conversation(
            source_id=source_id,
            inbox_id=inbox_id,
            contact_id=contact_id,
            custom_attributes=custom_attributes,
            message={
                "content": first_message
            } if first_message else None
        )
        return conversation

    async def touch_contact(self, contact_ref: ContactLead) -> Optional[Dict[str, Any]]:
        """ Touch a contact by identifier 
            If not found, create a new contact.
        """
        phone = format_to_e164(contact_ref.phone_number) if contact_ref.phone_number else None
        res = await self.client.search_contact(contact_ref.identifier)
        contacts = res.get("payload", [])

        if len(contacts) > 1:
            log.critical(f"Multiple contacts found for ref_identifier: {contact_ref.identifier} count: {len(contacts)}")
            # raise Exception(f"Multiple contacts found for identifier: {contact_ref.identifier}")
            

        contact = contacts[0] if contacts else None

        if not contact:
            log.debug(f"Contact with phone: {phone} and id:{contact_ref.identifier}  not found. Creating...")
            res = await self.client.create_contact(
                inbox_id=self.inbox.id,
                name=contact_ref.name,
                email=contact_ref.email,
                phone_number=phone,
                identifier=contact_ref.identifier,
                custom_attributes={
                    "celai_lead": contact_ref.celai_lead
                }
            )
            contact = res.get('payload', {}).get('contact')

            if not contact:
                raise Exception(f"Failed to create contact: {contact_ref} in Chatwoot: {self.base_url}. Message: {res.get('message')}")
        else: 
            # Update contact with custom attributes if not present
            cto_celai_lead = contact.get('custom_attributes', {}).get('celai_lead')
            if not cto_celai_lead or cto_celai_lead != contact_ref.celai_lead:
                log.debug(f"Updating contact: {contact_ref.identifier} with custom attributes")
                
                # Merge custom_attributes instead of overwriting
                merged_custom_attributes = dict(contact.get('custom_attributes', {}))
                merged_custom_attributes["celai_lead"] = contact_ref.celai_lead
                
                res = await self.client.update_contact(
                    contact_id=contact.get('id'),
                    inbox_id=self.inbox.id,
                    name=contact_ref.name,
                    email=contact_ref.email,
                    phone_number=phone,
                    identifier=contact_ref.identifier,
                    custom_attributes=merged_custom_attributes
                )
                contact = res.get('payload', {})
                # contact = await self.client.get_contact(contact.get('id'))
                if not contact:
                    raise Exception(f"Failed to update contact: {contact_ref} in Chatwoot: {self.base_url}. Message: {res.get('message')}")

        return contact

    async def touch_conversation(self, contact_ref: ContactLead) -> Optional[ChatwootConversationRef]:
        """ Touch a conversation by Contact identifier 
            If not found, create a new conversation.
        """

        conversation = await self.get_conversation(contact_ref.identifier)
        if not conversation:
            # Get or create contact
            contact = await self.touch_contact(contact_ref)

            cto_inboxes = contact.get("contact_inboxes", [])
            # Get source_id from contact_inboxe where inbox_id == self.inbox.id
            source_id = None
            for cto_inbox in cto_inboxes:
                if cto_inbox.get("inbox", {}).get("id") == self.inbox.id:
                    source_id = cto_inbox.get("source_id")
                    break

            if not source_id:
                raise Exception(f"Source ID not found for inbox: {self.inbox.id}")

            # Keep celai_lead in custom_attributes
            # in order to route future incoming messages from Chatwoot 
            # to the correct Celai channel
            conversation = await self.create_new_conversation(
                    contact_id=contact.get("id"), 
                    source_id=source_id,
                    custom_attributes={
                        "celai_lead": contact_ref.celai_lead
                    })

            if not conversation:
                raise Exception(f"Failed to create conversation for contact: {contact} in Chatwoot: {self.base_url}")

            log.debug(f"Conversation created: {conversation}")
            self.conversations[contact_ref.identifier] = ChatwootConversationRef(
                id=conversation.get("id"),
                identifier=contact_ref.identifier,
                # current time posix
                updated_at=int(time.time())
            )
            log.debug(f"Conversation touched: {self.conversations[contact_ref.identifier]}")
            return self.conversations[contact_ref.identifier]

        return conversation

    async def send_outgoing_text_message(self, contact_ref: ContactLead, message: str, private: bool = False) -> Optional[Dict[str, Any]]:
        """ Send an outgoing message to a contact 
            If contact not found, create a new contact.
        """
        try:           
            conversation = await self.touch_conversation(contact_ref)
            conversation_id = conversation.id
            return await self.client.create_message(
                account_id=self.account_id,
                conversation_id=conversation_id,
                content=message,
                private=private,
                content_type="text",
                message_type="outgoing",
                content_attributes={
                    "celai_generated": True,
                }
            )
        except Exception as e:
            log.error(f"Middleware Chatwoot: Error sending outgoing message: {e}")
            return None

    async def send_incoming_text_message(self, contact_ref: ContactLead, message: str, private: bool = False) -> Optional[Dict[str, Any]]:
        """ Send an incoming message to a contact 
            If contact not found, create a new contact.
        """
        
        try:
            conversation = await self.touch_conversation(contact_ref)
            conversation_id = conversation.id

            return await self.client.create_message(
                account_id=self.account_id,
                conversation_id=conversation_id,
                content=message,
                private=private,
                content_type="text",
                message_type="incoming"
            )
        except Exception as e:
            log.error(f"Middleware Chatwoot: Error sending incoming message: {e}")
            return None



from abc import ABC
from typing import Any
from deepgram import DeepgramClient, PrerecordedOptions, ClientOptionsFromEnv
from cel.voice.base_voice_provider import BaseVoiceProvider



class DeepgramAdapter(BaseVoiceProvider):
    def __init__(self, 
                 smart_format:bool = True, 
                 detect_language: bool = True, 
                 model: str = None, 
                 **kwargs):

        self.deepgram = DeepgramClient("", ClientOptionsFromEnv())

        self.options = PrerecordedOptions(
            model=model or "nova-2-general",
            smart_format=smart_format or True,
            detect_language=detect_language or True,
            **kwargs
        )

    async def STT(self, audio: bytes | str ) -> str:
        if isinstance(audio, bytes):
            payload = {"buffer": audio}
            response = await self.deepgram.listen.asyncprerecorded.v("1").transcribe_file(audio, self.options)
        if isinstance(audio, str):
            payload = {"url": audio}
            response = await self.deepgram.listen.asyncprerecorded.v("1").transcribe_url(payload, self.options)

        transcript = response["results"]["channels"][0]["alternatives"][0]["transcript"]
        return transcript
    
    async def TTS(self, text: str, voice: str = None, settings: Any = None):
        raise NotImplementedError




from .telegram_connector import TelegramConnector
from .model.telegram_lead import TelegramLead
from .model.telegram_message import TelegramMessage
from .model.telegram_attachment import TelegramAttachment


#   {
#     "type": "video",
#     "video": {
#       "link": "the-provider-name/protocol://the-url"
#       # provider is an optional parameter
#       "provider": {
#         "name" : "provider-name"
#       }
#     }
#   }

from .component import Component
from .utils import del_none
import json


class Video(Component):
    def __init__(self, link: str, provider: str | None = None) -> None:
        self.link = link
        self.provider = provider

    def __str__(self):
        video = {
            "type": "video",
            "video": {
                "link": self.link,
                "provider": {"name": self.provider} if self.provider else None
            }
        }
        del_none(video)
        return json.dumps(video)

    def __repr__(self):
        return f"VideoComponent(link={self.link}, provider={self.provider})"

    def __eq__(self, other):
        return self.link == other.link and self.provider == other.provider
    
    
    
    
    
if __name__ == "__main__":
    video = Video("the-provider-name/protocol://the-url", "provider-name")
    print(video)
    # Expected output: '{"type": "video", "video": {"link": "the-provider-name/protocol://the-url", "provider": {"name": "provider-name"}}}
    video = Video("the-provider-name/protocol://the-url")
    print(video)    
    # Expected output: '{"type": "video", "video": {"link": "the-provider-name/protocol://the-url", "provider": null}}'


from fastapi import APIRouter
from typing import Any, BinaryIO
from .message_gateway_context import MessageGatewayContext


class ConnectorsRegistry(type):
    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        if not hasattr(cls, '_instances'):
            cls._instances = []  # Crear solo en BaseLead

    def __call__(cls, *args, **kwargs):
        instance = super().__call__(*args, **kwargs)
        # Añadir instancia a BaseLead._instances
        BaseConnector._instances.append(instance)
        return instance
    
    def get_connector_by_name(cls, name):
        for instance in cls._instances:
            if instance.name() == name:
                return instance
        return None
    
    def get_all_connectors(cls):
        return cls._instances or []
    

class BaseConnector(metaclass=ConnectorsRegistry):

    def name(self) -> str:
        raise NotImplementedError

    def get_router(self) -> APIRouter:
        raise NotImplementedError

    def startup(self, context: MessageGatewayContext):
        raise NotImplementedError

    def shutdown(self, context: MessageGatewayContext):
        raise NotImplementedError

    def pause(self):
        raise NotImplementedError

    def resume(self):
        raise NotImplementedError

    def set_gateway(self, gateway):
        raise NotImplementedError

    async def send_text_message(self, lead, text: str, metadata: dict = {}, is_partial: bool = True):
        raise NotImplementedError
    
    async def send_voice_message(self, lead, text: str, options: dict = {}, is_partial: bool = False):
        raise NotImplementedError
    
    # Send an image message from memory
    async def send_image_message(self, lead, image: Any, filename:str, caption:str = None, metadata: dict = {}, is_partial: bool = True):
        raise NotImplementedError

    async def send_document_message(self, lead, content: str | bytes | BinaryIO, filename:str, caption:str = None, metadata: dict = {}):
        raise NotImplementedError
    
    async def send_image_url_message(self, lead, image_url: str, caption:str = None, metadata: dict = {}, is_partial: bool = True):
        raise NotImplementedError
    
    async def send_typing_action(self, lead):
        raise NotImplementedError
    
    async def send_message(self, message):
        raise NotImplementedError
    
    async def send_link_message(self, 
                            lead, 
                            text: str, 
                            links: list, 
                            metadata: dict = {}, 
                            is_partial: bool = True):
        raise NotImplementedError


import json
from loguru import logger as log
from cel.stores.history.base_history_provider import BaseHistoryProvider


class InMemoryHistoryProvider(BaseHistoryProvider):

    def __init__(self, key_prefix: str = "h"):
        self.store = {}
        self.key_prefix = key_prefix
        log.warning(f"Create: InMemoryHistoryProvider - Avoid using this in production.")

    def get_key(self, sessionId: str):
        return f"{self.key_prefix}:{sessionId}"

    async def append_to_history(self, sessionId: str, entry, metadata=None, ttl=None):
        key = self.get_key(sessionId)
        value = json.dumps(entry)
        if key not in self.store:
            self.store[key] = []
        self.store[key].append(value)

    async def get_history(self, sessionId: str):
        key = self.get_key(sessionId)
        values = self.store.get(key, [])
        res = [json.loads(v) for v in values]
        # remove None elements
        res = [r for r in res if r]

        return res

    async def clear_history(self, sessionId: str, keep_last_messages=None):
        key = self.get_key(sessionId)
        if keep_last_messages:
            self.store[key] = self.store[key][:keep_last_messages]
        else:
            self.store.pop(key, None)

    async def get_history_slice(self, sessionId: str, start, end):
        key = self.get_key(sessionId)
        history = self.store.get(key, [])[start:end]
        return [json.loads(h) for h in history]

    async def get_last_messages(self, sessionId: str, count):
        key = self.get_key(sessionId)
        history = self.store.get(key, [])[-count:]
        return [json.loads(h) for h in history]

    async def close_conversation(self, sessionId: str):
        raise NotImplementedError("Method not implemented.")


import json
from cel.stores.common.key_value_store import ListStore
from cel.stores.history.base_history_provider import BaseHistoryProvider



class RedisHistoryProviderAsync(BaseHistoryProvider):

    def __init__(self, store: ListStore, key_prefix: str = "h", ttl=None):
        """ Create a new RedisHistoryProviderAsync instance.
        :param store: Redis store
        :param key_prefix: Prefix for the keys
        :param ttl: Time to live in seconds for history. If None, it will never expire
        """
        
        print(f"Create: RedisHistoryProviderAsync")
        self.store = store
        self.key_prefix = key_prefix
        self.ttl = ttl

    def get_key(self, sessionId: str):
        return f"{self.key_prefix}:{sessionId}"

    async def append_to_history(self, sessionId: str, entry, metadata=None, ttl=None):
        key = self.get_key(sessionId)
        value = json.dumps(entry)
        await self.store.list_append(key, value, self.ttl or ttl)


    async def get_history(self, sessionId: str):
        key = self.get_key(sessionId)
        values = await self.store.list_get(key)
        res = [json.loads(v) for v in values]
        # remove None elements
        res = [r for r in res if r]
        return res
        

    async def clear_history(self, sessionId: str, keep_last_messages=None):
        key = self.get_key(sessionId)

        if keep_last_messages:
            # self.client.ltrim(key, 0, keep_last_messages)
            msgs = await self.store.list_get_last(key, keep_last_messages)
            await self.store.list_clear(key)
            for msg in msgs:
                await self.store.list_append(key, msg)
                
        else:
            await self.store.list_clear(key)


    async def get_last_messages(self, sessionId: str, count):
        key = self.get_key(sessionId)
        # history = self.client.lrange(key, -count, -1)
        history = await self.store.list_get_last(key, count)
        return [json.loads(h) for h in history]

    async def close_conversation(self, sessionId: str):
        raise NotImplementedError("Method not implemented.")
    

